"""
snapshot connector file for aws type snapshot. When the 'type' is 'aws', the
resource objects have to be fetched from the AWS interface.
The connection could use python SDK libraries or https calls or other known methods
provided by the provider.
Here in case of AWS, have used boto python SDK released and maintained by AWS.
boto library provides different 'describe_' functions and using this functionality,
the snapshot mentions which describe function has to be called, so 'type' could
be 'security_groups' , 'instances', 'regions' and using python reflection capabilities
a callable from describe_security_groups gets callable for this function to return security groups.
The describe_ functions lots of methods to query AWS resources, since we know what resource we need
we always query by id, so AWS snapshots shall always have {"id": "123de23"} to uniquely
identify the resource object.
"""
import json
import hashlib
import time
import pymongo
import os
import re
from googleapiclient import discovery
from oauth2client.service_account import ServiceAccountCredentials
from processor.helper.file.file_utils import exists_file
from processor.logging.log_handler import getlogger
from processor.helper.httpapi.http_utils import http_get_request
from processor.helper.config.rundata_utils import put_in_currentdata, get_dbtests, get_from_currentdata
from processor.helper.json.json_utils import get_field_value, json_from_file,\
    collectiontypes, STRUCTURE, save_json_to_file, get_field_value_with_default,\
    make_snapshots_dir, store_snapshot
from processor.connector.vault import get_vault_data
from processor.helper.config.config_utils import config_value, get_test_json_dir, framework_dir, CUSTOMER, EXCLUSION
from processor.database.database import insert_one_document, sort_field, get_documents,\
    COLLECTION, DATABASE, DBNAME, get_collection_size, create_indexes
from processor.helper.httpapi.restapi_azure import json_source
from processor.connector.snapshot_utils import validate_snapshot_nodes

logger = getlogger()
google_parameters = None

def get_google_parameters():
    """
    Return the google parameter object read from database or the filesystem
    """
    global google_parameters
    if not google_parameters:
        params_source = config_value('GOOGLE', 'params')
        if json_source():
            dbname = config_value(DATABASE, DBNAME)
            collection = config_value(DATABASE, collectiontypes[STRUCTURE])
            parts = params_source.rsplit('/')
            name = parts[-1].split('.')
            qry = {'name': name[0]}
            sort = [sort_field('timestamp', False)]
            docs = get_documents(collection, dbname=dbname, sort=sort, query=qry, limit=1)
            logger.info('Number of Google Params versions: %s', len(docs))
            if docs and len(docs):
                google_parameters = docs[0]['json']
        else:
            params_file = '%s/%s' % (framework_dir(), params_source)
            logger.info(params_file)
            if exists_file(params_file):
                google_parameters = json_from_file(params_file)
    return google_parameters

def get_google_data(snapshot_source):
    """
    The Google source object to be fetched from database or the filesystem
    The initial configuration for database is 'validator' and collection
    is 'structures', whereas for the filesystem the path to fetch the
    'structures' is  $SOLUTIONDIR/realm/<structure>.json
    """
    sub_data = {}
    if json_source():
        dbname = config_value(DATABASE, DBNAME)
        collection = config_value(DATABASE, collectiontypes[STRUCTURE])
        parts = snapshot_source.split('.')
        qry = {'name': parts[0]}
        sort = [sort_field('timestamp', False)]
        docs = get_documents(collection, dbname=dbname, sort=sort, query=qry, limit=1)
        logger.info('Number of Google structure Documents: %d', len(docs))
        if docs and len(docs):
            sub_data = docs[0]['json']
    else:
        json_test_dir = get_test_json_dir()
        file_name = '%s.json' % snapshot_source if snapshot_source and not \
            snapshot_source.endswith('.json') else snapshot_source
        google_source = '%s/../%s' % (json_test_dir, file_name)
        logger.info('Google source: %s', google_source)
        if exists_file(google_source):
            sub_data = json_from_file(google_source)

    if not sub_data:
        logger.error("Google connector file %s does not exist, or it does not contains the valid JSON.", snapshot_source)
    return sub_data



def generate_request_url(base_url, project_id):
    """Generate request url from base url"""
    try:
        path_list = base_url.split("https://")
        path_list = path_list[1].split('/')

        request_url = "https:/"
        for path in path_list:
            field_expresion = re.compile("\{([^}]+)\}")
            field = field_expresion.search(path)
            if field:
                field = path[1:len(path)-1]
                if field in ["projectId", "project"]:
                    path = project_id
                elif field in ["zone"]:
                    path = "-"
            request_url = "%s/%s" % (request_url, path)

        logger.info(request_url)
        return request_url
    except:
        logger.error("Invalid api url")
        return None

def get_api_path(node_type):
    """
    Get api path for populate snapshot
    """
    api_url = None
    params = get_google_parameters()
    if params and "GoogleApis" in params:
        for method, request_url in params['GoogleApis'].items():
            if method == node_type:
                api_url = request_url
    return api_url

def get_node(credentials, node, snapshot_source, snapshot):
    """
    Fetch node from google using connection. In this case using google client API's
    functions.
    """
    collection = node['collection'] if 'collection' in node else COLLECTION
    parts = snapshot_source.split('.')
    project_id = get_field_value_with_default(snapshot, 'project-id',"")
    path = get_field_value_with_default(node, 'path',"")
    zone = re.findall(r"(?<=zones\/)[a-zA-Z0-9\-]*(?=\/)", path)
    db_record = {
        "structure": "google",
        "error": None,
        "reference": project_id,
        "contentType": "json",
        "source": parts[0],
        "path": path,
        "timestamp": int(time.time() * 1000),
        "queryuser": get_field_value(snapshot, 'testUser'),
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "node": node,
        "snapshotId": node['snapshotId'],
        "collection": collection.replace('.', '').lower(),
        "region" : zone[0] if zone else "",
        "json": {}  # Refactor when node is absent it should None, when empty object put it as {}
    }

    try:
        access_token = credentials.get_access_token().access_token
        header = {
            "Authorization" : ("Bearer %s" % access_token)
        }

        node_type = node['type'] if node and 'type' in node else ""
        base_node_type_list = node_type.split("/")
        if len(base_node_type_list) > 1:
            base_node_type = base_node_type_list[0]
        else:
            logger.error("Invalid node type %s", node_type)
            return db_record
        
        base_url = "%s%s" % (base_node_type, ".googleapis.com")
        request_url = "https://%s/%s" % (base_url, path)
        logger.info("Invoke request for get snapshot: %s", request_url)
        status, data = http_get_request(request_url, header)
        logger.info('Get snapshot status: %s', status)
        
        if status and isinstance(status, int) and status == 200:
            if data:
                db_record['json'] = data
                checksum = get_checksum(data)
                if checksum:
                    db_record['checksum'] = checksum
        else:
            logger.error("Get snapshot returned invalid status: %s", status)
            db_record['error'] = ("Get snapshot returned invalid status: %s" % status)
    except Exception as ex:
        logger.error('Failed to populate the snapshot : %s', ex)
        db_record['error'] = 'Failed to populate the snapshot: %s' % ex
    
    return db_record


def get_all_nodes(credentials, node, snapshot_source, snapshot, snapshot_data):
    """
    Fetch all nodes from google using connection using google client API's functions.
    """
    collection = node['collection'] if 'collection' in node else COLLECTION
    parts = snapshot_source.split('.')
    project_id = get_field_value_with_default(snapshot, 'project-id',"")
    node_type = get_field_value_with_default(node, 'type',"")
    
    db_record = {
        "structure": "google",
        "error": None,
        "reference": project_id,
        "contentType": "json",
        "source": parts[0],
        "path": "",
        "timestamp": int(time.time() * 1000),
        "queryuser": get_field_value(snapshot, 'testUser'),
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "node": node,
        "snapshotId": None,
        "masterSnapshotId": [node['masterSnapshotId']],
        "collection": collection.replace('.', '').lower(),
        "json": {},  # Refactor when node is absent it should None, when empty object put it as {}
        "items": []
    }

    if node_type:
        access_token = credentials.get_access_token().access_token
        header = {
            "Authorization" : ("Bearer %s" % access_token)
        }

        base_node_type_list = node_type.split("/")
        if len(base_node_type_list) > 1:
            base_node_type = base_node_type_list[1]
        else:
            logger.error("Invalid node type '%s'", node_type)
            return db_record

        request_url = get_api_path(base_node_type)
        request_url = generate_request_url(request_url, project_id)
        logger.info("Invoke request for get snapshot: %s", request_url)
        
        status, data = http_get_request(request_url, header)
        logger.info('Get snapshot status: %s', status)

        fn_str_list = ""
        if node and 'type' in node and node['type']:
            fn_str_list = get_field_value(node, 'type').split(".")
        
        response_param = ""
        if fn_str_list and len(fn_str_list) > 1:
            response_param = fn_str_list[-2]
        elif fn_str_list and len(fn_str_list) == 1:
            response_param = fn_str_list[0]
        
        if data:
            check_node_type = node_type 
            node_type_list = node_type.split(".")
            if len(node_type_list) > 1:
                del node_type_list[-1]
                check_node_type = ".".join(node_type_list)

            db_record['json'] = data
            if response_param in data:
                db_record['items'] = data[response_param]
            elif "items" in data:
                if isinstance(data['items'], dict):
                    for name, scoped_dict in data['items'].items():
                        if response_param in scoped_dict:
                            db_record['items'] = db_record['items'] + scoped_dict[check_node_type]

                if not db_record['items']:
                    db_record['items'] = data['items']

            set_snapshot_data(node, db_record['items'], snapshot_data)

            checksum = get_checksum(data)
            if checksum:
                db_record['checksum'] = checksum

    return db_record

def set_snapshot_data(node, items, snapshot_data):
    if node['masterSnapshotId'] not in snapshot_data or not isinstance(snapshot_data[node['masterSnapshotId']], list):
        snapshot_data[node['masterSnapshotId']] =  []

    # create the node type for sub resources
    node_type = get_field_value(node, "type")
    node_type_list = node_type.split(".")
    resource_node_type = node_type
    if len(node_type_list) > 1:
        del node_type_list[-1]
        node_type_list.append("get")
        resource_node_type = ".".join(node_type_list)

    count = 0
    resource_items = []
    if isinstance(items, dict):
        for zone, resource in items.items():
            if 'selfLink' in resource:
                resource_items.append(resource)
            else:
                resource_type = node_type.split("/")[1].split(".")[-2]
                if resource_type in resource and isinstance(resource[resource_type], list):
                    if len(resource[resource_type]) > 0 and 'selfLink' in resource[resource_type][0]:
                        resource_items += resource[resource_type]
    else:
        resource_items = items

    includeSnapshotConfig = get_from_currentdata("INCLUDESNAPSHOTS")
    includeSnapshots = get_from_currentdata("SNAPHSHOTIDS")
    ignoreExclusion = False
    ignoreNode = False
    if includeSnapshotConfig:
        if node['masterSnapshotId'] in includeSnapshots:
            ignoreExclusion = True
            ignoreNode = False
        else:
            ignoreNode  = True

    exclusions = get_from_currentdata(EXCLUSION).get('exclusions', [])
    resourceExclusions = {}
    if not ignoreExclusion:
        for exclusion in exclusions:
            if 'exclusionType' in exclusion and exclusion['exclusionType'] and exclusion['exclusionType'] == 'resource':
                if 'paths' in exclusion and isinstance(exclusion['paths'], list):
                    resourceExclusions[tuple(exclusion['paths'])] = exclusion

    for item in resource_items:
        count += 1
        path_list = item['selfLink'].split("https://")
        path_list = path_list[1].split('/')
        path = "/".join(path_list[1:])

        found_old_record = False
        for masterSnapshotId, snapshot_list in snapshot_data.items():
            old_record = None
            if isinstance(snapshot_list, list):
                for item in snapshot_list:
                    if item["path"] == path:
                        old_record = item

                if old_record:
                    found_old_record = True
                    if node['masterSnapshotId'] not in old_record['masterSnapshotId']:
                        old_record['masterSnapshotId'].append(
                            node['masterSnapshotId'])

        if not found_old_record:
            if isinstance(path, str):
                key = tuple([path])
            else:
                key = None
            if key and key in resourceExclusions:
                logger.warning("Excluded from resource exclusions: %s", path)
                continue
            if not ignoreNode:
                snapshot_data[node['masterSnapshotId']].append(
                    {
                        "masterSnapshotId" : [node['masterSnapshotId']],
                        "snapshotId": '%s%s' % (node['masterSnapshotId'], str(count)),
                        "type": resource_node_type,
                        "collection": node['collection'],
                        "path": path,
                        "status" : "active",
                        "validate" : node['validate'] if 'validate' in node else True
                    })
    return snapshot_data

def get_checksum(data):
    """ Get the checksum for the Google data fetched."""
    checksum = None
    try:
        data_str = json.dumps(data)
        checksum = hashlib.md5(data_str.encode('utf-8')).hexdigest()
    except:
        pass
    return checksum


def populate_google_snapshot(snapshot, container=None):
    """
    This is an entrypoint for populating a snapshot of type google.
    All snapshot connectors should take snapshot object and based on
    'source' field create a method to connect to the service for the
    connector.
    The 'source' field could be used by more than one snapshot, so the
    'testuser' attribute should match to the user the 'source'
    """
    dbname = config_value('MONGODB', 'dbname')
    snapshot_source = get_field_value(snapshot, 'source')
    snapshot_user = get_field_value(snapshot, 'testUser')
    project_id = get_field_value(snapshot, 'project-id')
    sub_data = get_google_data(snapshot_source)
    snapshot_nodes = get_field_value(snapshot, 'nodes')
    snapshot_data, valid_snapshotids = validate_snapshot_nodes(snapshot_nodes)
    if valid_snapshotids and sub_data and snapshot_nodes:
        logger.debug(sub_data)
        try:
            for node in snapshot['nodes']:
                validate = node['validate'] if 'validate' in node else True
                logger.info(node)
                node_type = get_field_value_with_default(node, 'type',"")
                credentials = get_google_client_data(sub_data, snapshot_user, node_type, project_id)
                if not credentials:
                    logger.info("No  GCE connection in the snapshot to access Google resource!...")
                    return snapshot_data
                if 'snapshotId' in node:
                    if validate:
                        data = get_node(credentials, node, snapshot_source, snapshot)
                        if data:
                            error_str = data.pop('error', None)
                            if get_dbtests():
                                if get_collection_size(data['collection']) == 0:
                                    #Creating indexes for collection
                                    create_indexes(
                                        data['collection'], 
                                        config_value(DATABASE, DBNAME), 
                                        [
                                            ('snapshotId', pymongo.ASCENDING),
                                            ('timestamp', pymongo.DESCENDING)
                                        ]
                                    )
                                    create_indexes(
                                        data['collection'], 
                                        config_value(DATABASE, DBNAME), 
                                        [
                                            ('_id', pymongo.DESCENDING),
                                            ('timestamp', pymongo.DESCENDING),
                                            ('snapshotId', pymongo.ASCENDING)
                                        ]
                                    )
                                insert_one_document(data, data['collection'], dbname)
                            else:
                                snapshot_dir = make_snapshots_dir(container)
                                if snapshot_dir:
                                    store_snapshot(snapshot_dir, data)
                    
                            if 'masterSnapshotId' in node:
                                snapshot_data[node['snapshotId']] = node['masterSnapshotId']
                            else:
                                snapshot_data[node['snapshotId']] = False if error_str else True
                        else:
                            node['status'] = 'inactive'
                elif 'masterSnapshotId' in node:
                        data = get_all_nodes(credentials, node, snapshot_source, snapshot, snapshot_data)
                        logger.debug('Type: %s', type(data))
        except Exception as ex:
            logger.info('Unable to create Google client: %s', ex)
            raise ex
    return snapshot_data



def generate_gce(google_data, project, user):
    """
    Generate client secret json from the google data
    """
    logger.info("Generating GCE")
    gce = {
        "type": get_field_value(user, "type"),
        "project_id": get_field_value(project, "project-id"),
        "private_key_id": get_field_value(user, "private_key_id"),
        "private_key": get_field_value(user, "private_key"),
        "client_email": get_field_value(user, "client_email"),
        "client_id": get_field_value(user, "client_id"),
        "auth_uri": get_field_value(google_data, "auth_uri"),
        "token_uri": get_field_value(google_data, "token_uri"),
        "auth_provider_x509_cert_url": get_field_value(google_data, "auth_provider_x509_cert_url"),
        "client_x509_cert_url": get_field_value(user, "client_x509_cert_url"),
    }

    # Read the private key from the key path
    if not gce['private_key'] and get_field_value(user, "private_key_path"):
        private_key_path = get_field_value(user, "private_key_path")
        logger.info("Private key path : %s ", private_key_path)
        try:
            gce['private_key'] = open(private_key_path, 'r', encoding="utf-8").read().replace("\\n","\n")
            if gce['private_key']:
                logger.info('Private key from Private key path, Secret: %s', '*' * len(gce['private_key']))
        except Exception as e:
            raise Exception("Private key does not exist at given private key path : %s " % str(e))
        
        if not gce['private_key']:
            raise Exception("Private key does not exist at given private key path : %s " % private_key_path)
    
    # Read the private key from the vault
    if not gce['private_key']:
        private_key = get_vault_data(gce['private_key_id'])
        if private_key:
            gce["private_key"] = private_key.replace("\\n","\n")
            if gce["private_key"]:
                logger.info('Private key from vault Secret: %s', '*' * len(gce["private_key"]))
        elif get_from_currentdata(CUSTOMER):
            raise Exception("Private key does not set in a vault")
    
    if not gce['private_key']:
        raise Exception("No `private_key` field in the connector file to access Google resource!...")

    if (None in list(gce.values())):
        raise Exception("Connector file does not contains valid values or some fields are missing for access Google resources.")

    return gce


def get_google_client_data(google_data, snapshot_user, node_type, project_id):
    """
    Generate Google Service Account credentials object from the google structure file.
    """
    credentials = None
    found = False
    if google_data and snapshot_user:
        projects = get_field_value(google_data, "projects")
        for project in projects:
            structure_project_id = get_field_value(project, 'project-id')
            if structure_project_id == project_id:
                users = get_field_value(project, 'users')
                if users:
                    for user in users:
                        user_name = get_field_value(user, 'name')
                        if user_name == snapshot_user:
                            found = True
                            gce = generate_gce(google_data, project, user)
                            if gce:
                                save_json_to_file(gce, '/tmp/gce.json')
                                logger.info("Creating credential object")
                                scopes = ['https://www.googleapis.com/auth/compute', "https://www.googleapis.com/auth/cloud-platform"]
                                credentials = ServiceAccountCredentials.from_json_keyfile_name('/tmp/gce.json', scopes)
                                # service_name = get_service_name(node_type)
                                # compute = discovery.build(service_name, 'v1', credentials=credentials, cache_discovery=False)
                            break 
            if found:
                break
    return credentials



"""
   http utils using urllib default packaged with python3.
"""
import json
import copy
from urllib import request, parse
from urllib.error import HTTPError, URLError
from processor.helper.json.json_utils import json_from_string
from processor.logging.log_handler import getlogger
from processor.helper.config.rundata_utils import put_in_currentdata

logger = getlogger()
hdrs = {
    "Cache-Control": "no-cache",
    "Accept": "application/json"
}


def check_and_add_error(status, errmsg):
    """Add error based on the status code of the http response"""
    if status and isinstance(status, int) and status != 200:
        if status >= 400:
            put_in_currentdata('errors', errmsg)
        logger.info(errmsg)


def get_request_headers(headers=None):
    """Add json and no cache headers to the existing headers if passed."""
    req_headers = copy.copy(headers) if headers else {}
    req_headers.update(hdrs)
    return req_headers


def urlopen_request(urlreq, method):
    """Common utility to trigger the http request."""
    try:
        urlresp = request.urlopen(urlreq)
        respdata = urlresp.read()
        st_code = urlresp.status
        logger.debug("%s status: %d, response: %s", method, st_code, respdata)
        check_and_add_error(st_code, "%s Status: %d" % (method, st_code))
        if isinstance(respdata, bytes):
            respdata = respdata.decode()
        data = json_from_string(respdata)
        logger.debug("%s status: %d", method, st_code)
    except HTTPError as ex:
        # st_code = ex.code if method == "POST" else None
        st_code = ex.code
        data = ex.msg if method == "POST" else None
        logger.info("HTTP %s: status: %s, ex:%s ", method, st_code, ex)
    except URLError as ex:
        st_code = 500
        data = str(ex)
        logger.info("HTTP %s: status: %s, ex:%s ", method, st_code, ex)
    return st_code, data


def http_delete_request(url, deldata=None, headers=None, name='DELETE'):
    """Delete method sends and accepts JSON format with basic authentication"""
    logger.info("HTTP %s %s  .......", name, url)
    if not url:
        return None, None
    if deldata:
        encdata = parse.urlencode(deldata).encode()
        logger.info('%s: data: %s', name, encdata)
        urlreq = request.Request(url, data=encdata, headers=get_request_headers(headers),
                                 method=name)
    else:
        urlreq = request.Request(url, headers=get_request_headers(headers), method=name)
    return urlopen_request(urlreq, name)


def http_get_request(url, headers=None, name='HTTP GET'):
    """Get method sends and accepts JSON format."""
    logger.info("%s %s  .......", name, url)
    if not url:
        return None, None
    urlreq = request.Request(url, headers=get_request_headers(headers), method='GET')
    return urlopen_request(urlreq, name)


def http_put_request(url, mapdata, headers=None, name='PUT', json_type=False):
    """Put method sends and accepts JSON format."""
    logger.info("HTTP %s %s  .......", name, url)
    if not url:
        return None, None
    if not json_type:
        putdata = parse.urlencode(mapdata).encode()
    else:
        putdata = json.dumps(mapdata, cls=json.JSONEncoder).encode('utf-8')
    logger.info('%s: data: %s', name, putdata)
    urlreq = request.Request(url, data=putdata, headers=get_request_headers(headers),
                             method='PUT')
    return urlopen_request(urlreq, name)


def http_post_request(url, mapdata, headers=None, json_type=False, name='HTTP POST:'):
    """Post method sends and accepts JSON format"""
    logger.info("%s %s  .......", name, url)
    if not url:
        return None, None
    myhdrs = get_request_headers(headers)
    if json_type:
        myhdrs['Content-Type'] = 'application/x-www-form-urlencoded'
        postdata = parse.urlencode(mapdata).encode()
    else:
        postdata = parse.urlencode(mapdata).encode()
    logger.debug('%s: data: %s', name, postdata)
    urlreq = request.Request(url, data=postdata, headers=myhdrs,
                             method='POST')
    return urlopen_request(urlreq, name)


def http_json_post_request(url, mapdata, headers=None, name='HTTP POST:'):
    """Post method sends and accepts JSON format"""
    logger.info("%s %s  .......", name, url)
    if not url:
        return None, None
    myhdrs = get_request_headers(headers)
    postdata = json.dumps(mapdata)
    logger.debug('%s: data: %s', name, postdata)
    urlreq = request.Request(url, data=postdata.encode(), headers=myhdrs,
                             method='POST')
    return urlopen_request(urlreq, name)



import re
import time
from typing import Tuple
from processor.database.database import (
    get_documents,
    find_and_update_document,
    insert_one_document,
)
from processor.helper.config.config_utils import DATABASE, DBNAME, config_value
from processor.logging.log_handler import getlogger


logger = getlogger()


def get_collection_config_data(container: str) -> dict:
    # get config of given collection
    config_data = get_collection_config(container)
    collection_config_data = config_data.get("json", {}).get("configuration", {})
    return collection_config_data


def get_collection_config(container: str) -> dict:
    # get config of given collection
    dbname = config_value(DATABASE, DBNAME)

    collection = "structures"
    query = {"type": "collection_configuration", "container": container}
    config_data_list = get_documents(collection, query, dbname)

    return config_data_list[0] if config_data_list else {}


def update_collection_config(container: str, updated_json: dict) -> bool:
    dbname = config_value(DATABASE, DBNAME)
    collection = "structures"
    query = {"type": "collection_configuration", "container": container}
    updated_json = {"$set": updated_json}
    updated = find_and_update_document(collection, dbname, query, updated_json)
    return updated


def create_collection_config(container: str, configuration: dict):
    dbname = config_value(DATABASE, DBNAME)
    collection = "structures"
    structure = {
        "type": "collection_configuration",
        "collection": collection,
        "container": container,
        "timestamp": int(time.time() * 1000),
        "json": {"configuration": configuration},
    }
    doc_id_str = insert_one_document(structure, collection, dbname, False)
    return doc_id_str


def get_master_collection_config() -> dict:
    dbname = config_value(DATABASE, DBNAME)
    collection = "structures"
    query = {"type": "master_collection_configuration"}
    config_data_list = get_documents(collection, query, dbname)
    config_data = (
        config_data_list[0].get("json", {}).get("configurations", {})
        if config_data_list
        else {}
    )
    return config_data


def validate_master_config(
    collection_config, master_collection_config
) -> Tuple[bool, list]:
    is_valid = True
    messages = []
    if isinstance(master_collection_config, list):
        for master_config_item in master_collection_config:
            if isinstance(master_config_item, dict) and isinstance(
                collection_config, dict
            ):
                master_config_key = master_config_item.get("key")
                required = master_config_item.get("required", False)
                master_config_children = master_config_item.get("children", [])
                allowed_values = master_config_item.get("allowed_values", [])

                if (
                    master_config_key not in collection_config.keys()
                    and required is True
                ):
                    is_valid = False
                    message = f"'{master_config_key}' key is required"
                    messages.append(message)

                elif master_config_key in collection_config.keys():
                    collection_config_value = collection_config[master_config_key]
                    if master_config_children:
                        if isinstance(collection_config_value, dict):
                            is_valid, new_messages = validate_master_config(
                                collection_config_value, master_config_children
                            )
                            messages.extend(new_messages)
                        else:
                            pass
                    elif (
                        allowed_values and collection_config_value not in allowed_values
                    ):
                        is_valid = False
                        message = f"'{master_config_key}'s value '{collection_config_value}' is not a valid value, allowed valued are: {allowed_values}"
                        messages.append(message)

    return is_valid, messages


# master_config = [
#     {
#         "key": "generate_pr",
#         "allowed_values": [True, False],
#         "default": True,
#         "title": "Generate PR",
#         "description": "make it False to not to generate auto Pull Request",
#         "type": "bool",
#         "required": True,
#     },
#     {
#         "key": "aws_secret_remediation",
#         "title": "AWS Secret Remediation",
#         "description": "configure secret vault to store leaked secrets",
#         "type": "json",
#         "required": False,
#         "children": [
#             {
#                 "key": "account_id",
#                 "title": "AWS Account ID",
#                 "description": "AWS Cloud 12-digit account number",
#                 "type": "integer",
#                 "required": True,
#             },
#             {
#                 "key": "region",
#                 "title": "Region",
#                 "allowed_values": [
#                     "us-east-1",
#                     "us-east-2",
#                     "us-west-1",
#                     "us-west-2",
#                     "af-south-1",
#                     "ap-east-1",
#                     "ap-south-1",
#                     "ap-northeast-3",
#                     "ap-northeast-2",
#                     "ap-southeast-1",
#                     "ap-southeast-2",
#                     "ap-northeast-1",
#                     "ca-central-1",
#                     "eu-central-1",
#                     "eu-west-1",
#                     "eu-west-2",
#                     "eu-south-1",
#                     "eu-west-3",
#                     "eu-north-1",
#                     "me-south-1",
#                     "sa-east-1",
#                     "us-gov-east-1",
#                     "us-gov-west-1",
#                 ],
#                 "required": True,
#                 "Description": "Select a region where your secret manager is available",
#             },
#         ],
#     },
# ]
# test_config = {
#     "generate_pr": "False",
#     "aws_secret_remediation": {
#         "account_id": 123456789012,
#         "region": "us"
#     },
# }

# is_valid, messages = validate_master_config(test_config, master_config)

# if is_valid is False:
#     for message in messages:
#         print(message)
# else:
#     print("config file is valid")



"""all the base functions for making REST API calls"""
import json
import requests


jsonhdr = {
    "Content-Type": "application/json",
    "Accept": "application/json"
}


def json_delete_request(url, deldata=None, headers=None, log=False):
    "Delete method sends and accepts JSON format without basic authenication"
    if log: print("Deleting %s  ......." % url)
    data = None
    st_code = None
    if headers:
        headers.update(jsonhdr)
    else:
        headers = jsonhdr
    if url: #Do something only valid URL
        if deldata:
            resp = requests.delete(url, data=json.dumps(deldata), headers=headers)
        else:
            resp = requests.delete(url, headers=headers)
        if log: print("Get response: %s" % resp)
        st_code = resp.status_code
        try:
            data = resp.json()
        except:
            pass # Can we do anything here, not anything i can think of immediately
    else:
        pass # Do nothing.
    return st_code, data


def json_get_request(url, headers=None, log=False):
    "Get method sends and accepts JSON format without basic authenication"
    if log: print("Getting %s  ......." % url)
    data = None
    st_code = None
    if headers:
        headers.update(jsonhdr)
    else:
        headers = jsonhdr
    if url: #Do something only valid URL
        resp = requests.get(url, headers=headers)
        if log: print("Get response: %s" % resp)
        st_code = resp.status_code
        try:
            data = resp.json()
        except:
            pass # Can we do anything here, not anything i can think of immediately
    else:
        pass # Do nothing.
    return st_code, data


def json_put_request(url, mapdata, headers=None, log=False):
    "Put method sends and accepts JSON format without basic authenication"
    if log: print("Putting %s  ......." % url)
    data = None
    st_code = None
    if headers:
        headers.update(jsonhdr)
    else:
        headers = jsonhdr
    if url: #Do something only valid URL
        resp = requests.put(url, data=json.dumps(mapdata), headers=headers)
        if log: print("Get response: %s" % resp)
        st_code = resp.status_code
        try:
            data = resp.json()
        except:
            pass # Can we do anything here, not anything i can think of immediately
    else:
        pass # Do nothing.
    return st_code, data


def json_post_request(url, mapdata, headers=None, log=False):
    "Post method sends and accepts JSON format without basic authenication"
    if log: print("Posting %s  ......." % url)
    data = None
    st_code = None
    if headers:
        headers.update(jsonhdr)
    else:
        headers = jsonhdr
    if url: #Do something only valid URL
        resp = requests.post(url, data=json.dumps(mapdata), headers=headers)
        if log: print("Get response: %s" % resp)
        st_code = resp.status_code
        try:
            data = resp.json()
        except:
            pass # Can we do anything here, not anything i can think of immediately
    else:
        pass # Do nothing.
    return st_code, data



from processor.template_processor.base.base_template_processor import TemplateProcessor
from processor.helper.file.file_utils import exists_file,exists_dir
from processor.helper.yaml.yaml_utils import yaml_from_file,HelmChartConvertionKey
from yaml.loader import FullLoader
from processor.logging.log_handler import getlogger
from processor.templates.helm.helm_parser import HelmTemplateParser

logger = getlogger()

class HelmChartTemplateProcessor(TemplateProcessor):
    """
    For process helm charts
    """
    def __init__(self, node, **kwargs):
        super().__init__(node, tosave=False, **kwargs)
    
    def is_template_file(self, file_path):
        """
        check for valid template file for parse helm template
        """
        file_type = file_path.split(".")[-1]
        file_name = file_path.split("/")[-1].split(".")[0]
        if file_type == "yaml" and file_name == "Chart"  or HelmChartConvertionKey in file_path:
            helm_source = file_path.rpartition("/")[0]
            helm_template = HelmTemplateParser(helm_source)
            if helm_template.validate(helm_source):
                return True
            # file_path.rpartition("/")[0]
            return True
        return False

    def process_template(self, paths):
        """
        process the files stored at specified paths and returns the template
        """
        template_json = None
        
        if paths and isinstance(paths, list):
            template_file_path = ""
            # paths[0] = paths.
            for path in paths:
                file_path = '%s/%s' % (self.dir_path, path)
                logger.info("Fetching data : %s ", path)
                if self.is_template_file(file_path):
                    template_file_path = file_path

            self.template_file = template_file_path
            if template_file_path:
                template_json = yaml_from_file(file_path,loader=FullLoader)
                if template_json:
                    self.contentType = 'yaml'
                    if template_json.get("kind"):
                        self.resource_types = [template_json.get("kind").lower()]
                self.contentType = 'yaml'
        return template_json
        
        
        



from typing import List
from hcl2.transformer import DictTransformer

class HClDictTransformer(DictTransformer):
    
    def full_splat(self, args: List) -> str:
        return ".".join(args)
    
    def full_splat_expr_term(self, args: List) -> str:
        return "%s[*].%s" % (args[0], args[1])




COMPLIANCES = [{
    "masterTestId":"SENSITIVE_EXTENSION_TEST",
    "masterSnapshotId" : [
        "ALL"
    ],
    "type":"python",
    "rule":"file(sensitive_extension.py)",
    "evals":[
        {
            "id":"PR-COM-SEN-EXT-001",
            "eval":"data.rule.sensitive_extensions",
            "message":"data.rule.sensitive_extensions_err",
            "remediationDescription":"you need to add these extensions to your .gitignore file to prevent them from checking in to your repository. these files need to be moved securely to a vault to be managed securely and then referenced in your code",
            "remediationFunction":""
        }
    ],
    "severity":"Medium",
    "title":"Sensitive files should not be checked into the git repo",
    "description":"Certain file types contain sensitive information and should not be checked into the git repositories. You need to move these files to a vault and reference them from your code. Prancer checks for the following file types to make sure they are not in the repo:<br>*.PFX or *.P12 - Personal Information Exchange Format<br>*.PEM - a Base64 encoded DER certificate<br>*.CER or *.CRT - Base64-encoded or DER-encoded binary X.509 Certificate<br>*.CRL - Certificate Revocation List<br>*.CSR - Certificate Signing Request<br>*.DER - DER-encoded binary X.509 Certificate<br>*.P7B or *.P7R or *.SPC - Cryptographic Message Syntax Standard<br>.Key – key files",
    "tags":[
        {
            "cloud":"git",
            "compliance":[
                "Best Practice"
            ],
            "service":[
                "common"
            ]
        }
    ],
    "resourceTypes":[
        "sensitive_extension"
    ],
    "status":"enable"
}]


import os
import shutil
import requests
import urllib.parse
from processor.helper.json.json_utils import save_json_to_file, json_from_file
from processor.helper.config.config_utils import config_value, framework_dir
from processor.helper.file.file_utils import mkdir_path, remove_file, exists_file
from processor.logging.log_handler import getlogger

LOCAL='LOCAL'
DEV='DEV'
QA='QA'
PROD='PROD'
SERVERENVS = [LOCAL, DEV, QA, PROD]

LOCALSERVER='http://localhost:8080/api/'
SERVERS = {
    DEV: 'portaldev',
    QA: 'portalqa',
    PROD: 'portal'
}
COLLECTIONAPI = 'vs/container/'
VALIDATIONAPI = 'validate/accesstoken/'

logger = getlogger()

def get_company_prefix(company):
    company_prefix = company
    if company and not company.startswith('prancer-'):
        company_prefix = 'prancer-%s' % company
    return company_prefix


def get_api_server(server, company):
    apiserver = None
    if server:
        if server in SERVERENVS:
            if server == LOCAL:
                apiserver = LOCALSERVER
            else:
                if company:
                    company_prefix = get_company_prefix(company)
                    apiserver = 'https://%s.prancer.io/%s/api/' % (SERVERS[server], company_prefix)
    return apiserver

def get_validate_token_api(apiserver):
    validationApi = None
    if apiserver:
        validationApi = apiserver + VALIDATIONAPI
    return validationApi

def get_collection_api(apiserver, collection=None):
    collectionApi = None
    if apiserver:
        collectionApi = apiserver + COLLECTIONAPI
        if collection:
         collectionApi = collectionApi + '?collection=' + urllib.parse.quote_plus(collection)
    return collectionApi


def create_container_compliance(container, data):
    """ Create mastersnapshot, mastertest and connectors for the container in the framework directory"""
    validFolder = config_value('TESTS', 'containerFolder')
    validpath = '%s/%s/%s' % (framework_dir(), validFolder, container)
    strFolder = config_value('AZURE', 'azureStructureFolder')
    if strFolder:
        strpath = '%s/%s' % (framework_dir(), strFolder)
        mkdir_path(strpath)
    else:
        strpath = '%s/%s/..' % (framework_dir(), validFolder)
    shutil.rmtree(validpath, ignore_errors=True)
    mkdir_path(validpath)
    outputpath = None
    for connector in data['connectors']:
        cname = '%s/%s.json' % (strpath, connector['name'])
        remove_file(cname)
        save_json_to_file(connector['json'], cname)
    for mastersnapshot in data['masersnapshots']:
        cname = '%s/%s.json' % (validpath, mastersnapshot['name'])
        remove_file(cname)
        save_json_to_file(mastersnapshot['json'], cname)
    for mastertest in data['mastertests']:
        cname = '%s/%s.json' % (validpath, mastertest['name'])
        outputpath = '%s/output-%s.json' % (validpath, mastertest['name'])
        remove_file(cname)
        save_json_to_file(mastertest['json'], cname)
    for exclusion in data['exclusions']:
        cname = '%s/exclusions.json' % validpath
        remove_file(cname)
        save_json_to_file(exclusion, cname)
    return outputpath

# def upload_compliance_results(container, opath, server, company, apitoken):
def upload_compliance_results_multipart(container, opath, server, company, apitoken):
    from processor.logging.log_handler import get_dblog_name, FWLOGFILENAME, FWLOGGER
    dlog = get_dblog_name()
    fname = FWLOGFILENAME
    name = fname.rsplit('/', 1)
    oname = opath.rsplit('/', 1)
    apiserver = get_api_server(server, company)
    if apiserver:
        collectionUri = get_collection_api(apiserver)
        hdrs = {
            "Authorization": "Bearer %s" % apitoken
        }
        files = (
            ('collection', (None, container)),
            ('log', ('logs_' + name[-1], open(fname, 'r'), 'application/octet-stream', {'Expires': '0', 'chunk': 0})),
            ('output', (oname[-1], open(opath, 'r'), 'application/octet-stream', {'Expires': '0'}))
        )
        resp = requests.post(collectionUri, headers=hdrs, files=files)
        if resp.status_code == 200:
            logger.info(resp.json())

def read_in_chunks(file_object, chunk_size=65536):
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data


def upload_file(container, content_path, content_name, url, logname, fileType, hdrs):
    # content_name = str(file)
    # content_path = os.path.abspath(file)
    content_size = os.stat(content_path).st_size

    # logger.info(content_name, content_path, content_size)

    f = open(content_path)

    index = 0
    offset = 0
    headers = hdrs
    # paths = content_name.rsplit('/', 1)

    data = read_in_chunks(f)
    chunk = next(data)
    moredata = True
    while True:
        offset = index + len(chunk)
        headers['Content-Type'] = 'application/octet-stream'
        headers['Content-length'] = str(content_size)
        nextdata = None
        data = read_in_chunks(f)
        try:
            nextdata = next(data)
        except StopIteration as ex:
            moredata = False

        # Content-Range'] = '<filename>:<offset start>:<offset end>:<0:file end, 1: file more> :<log|output|snapshot> :<container>:<logname>
        headers['Content-Range'] = '%s:%s:%s:%d:%s:%s:%s' % (content_name, index, offset,int(moredata), fileType, container, logname)
        index = offset
        try:
            r = requests.post(url, data=chunk.encode('utf-8'), headers=headers)
            # print("r: %s, Content-Range: %s" % (r, headers['Content-Range']))
        except Exception as e:
            print(e)
        if not moredata:
            break
        chunk = nextdata

# def upload_compliance_results_chunk(container, opath, server, company, apitoken):
def upload_compliance_results(container, opath, server, company, apitoken):
    from processor.logging.log_handler import get_dblog_name, FWLOGFILENAME, FWLOGGER
    fname = FWLOGFILENAME
    name = fname.rsplit('/', 1)
    logs = name[-1].split('.')
    oname = opath.rsplit('/', 1)
    ojson = json_from_file(opath)
    snapshotpath = None
    snapshot = None
    if ojson and 'snapshot' in ojson:
        snapshot = '%s.json' % ojson['snapshot']
        snapshotpath = os.path.join('', oname[0], snapshot)
        ojson = None
    apiserver = get_api_server(server, company)
    if apiserver:
        collectionUri = get_collection_api(apiserver)
        hdrs = {
            "Authorization": "Bearer %s" % apitoken
        }
        if exists_file(fname):
            upload_file(container, fname, name[-1], collectionUri, logs[0], 'log', hdrs)
        if exists_file(opath):
            upload_file(container, opath, oname[-1], collectionUri, logs[0], 'output', hdrs)
        if snapshot and snapshotpath and exists_file(snapshotpath):
            upload_file(container, snapshotpath, snapshot, collectionUri, logs[0], 'snapshot', hdrs)


import json
from processor.logging.log_handler import getlogger
from processor.helper.json.json_utils import json_from_file
from processor.template_processor.base.base_template_processor import TemplateProcessor
from processor.templates.aws.aws_parser import AWSTemplateParser, allowed_extensions
from processor.helper.file.file_utils import exists_file
from cfn_flip import to_json

logger = getlogger()

class AWSTemplateProcessor(TemplateProcessor):
    """
    Base Template Processor for process template 
    """
    def __init__(self, node, **kwargs):
        super().__init__(node, tosave=False, **kwargs)

    def is_parameter_file(self, file_path):
        """
        check for valid parameter file for parse cloudformation template
        """
        file_parts = file_path.split(".")
        if len(file_parts) > 0 and file_parts[-1] in allowed_extensions:
            json_data = json_from_file(file_path)
            if json_data and isinstance(json_data, list):
                parameter = json_data[0]
                if isinstance(parameter, dict) and "ParameterKey" in parameter and "ParameterValue" in parameter:
                    return True
        return False

    def is_template_file(self, file_path):
        """
        check for valid template file for parse cloudformation template
        """
        file_parts = file_path.split(".")
        if len(file_parts) > 0 and file_parts[-1] in allowed_extensions:
            template_json = None
            if file_path.endswith(".yaml") and exists_file(file_path):
                with open(file_path, encoding="utf-8") as yml_file:
                    try:
                        template_json = json.loads(to_json(yml_file.read()))
                        self.contentType = 'yaml'
                    except:
                        pass
            elif file_path.endswith(".json"):
                template_json = json_from_file(file_path)
                self.contentType = 'json'
            
            elif file_path.endswith(".template") or file_path.endswith(".txt"):
                template_json = json_from_file(file_path)
                if template_json:
                    self.contentType = 'json'
                else:
                    with open(file_path, encoding="utf-8") as yml_file:
                        try:
                            template_json = json.loads(to_json(yml_file.read()))
                            self.contentType = 'yaml'
                        except:
                            pass

            if template_json and "AWSTemplateFormatVersion" in template_json:
                return True
        return False

    def is_valid_file_extension(self, file_path):
        for extn in allowed_extensions:
            if file_path.endswith(extn):
                return True
        return False

    def process_template(self, paths):
        """
        process the files stored at specified paths and returns the template
        """
        template_json = None
        if paths and isinstance(paths, list):
            parameter_file = None
            template_file = None
            for path in paths:
                file_path = ('%s/%s' % (self.dir_path, path)).replace("//", "/")
                if self.is_valid_file_extension(file_path) and self.is_parameter_file(file_path):
                    parameter_file = file_path
                else:
                    template_file = file_path

            if template_file:
                self.template_files = [template_file]
                if parameter_file:
                    self.parameter_files = [parameter_file]
                    aws_template_parser = AWSTemplateParser(template_file, parameter_file=parameter_file)
                else:
                    aws_template_parser = AWSTemplateParser(template_file)
                template_json = aws_template_parser.parse()
                self.contentType = aws_template_parser.contentType
                self.resource_types = aws_template_parser.resource_types

        return template_json



import json
import ast
import sys
import hashlib
import time
import os
import pymongo
from datetime import datetime
from processor.helper.config.rundata_utils import get_dbtests, get_from_currentdata
from processor.helper.json.json_utils import get_field_value,json_from_file,\
    make_snapshots_dir,store_snapshot,get_field_value_with_default,STRUCTURE,\
    collectiontypes
from processor.logging.log_handler import getlogger
from processor.helper.config.config_utils import config_value,framework_dir, EXCLUSION
from kubernetes import client,config
import kubernetes.client
from processor.connector.snapshot_utils import validate_snapshot_nodes
from processor.database.database import insert_one_document,\
    COLLECTION, DATABASE, DBNAME, get_collection_size, create_indexes,sort_field,\
    get_documents
import traceback
from processor.helper.httpapi.restapi_azure import json_source
from processor.connector.snapshot_base import Snapshot
from processor.connector.vault import get_vault_data

Cache_secret = ""
Cache_namespace = ""


logger = getlogger()



def get_kubernetes_structure_path(snapshot_source):
    """
    get_kubernetes_structure_path will get kubernetes connector file path
    from configuration file.
    """
    folder = config_value('KUBERNETES','kubernetesStructureFolder')
    if folder:
        connector_path = '%s/%s/%s.json' % (framework_dir(), folder, snapshot_source)
    else:
        connector_path = '%s/%s.json' % (framework_dir(), snapshot_source)
    return connector_path

def get_kubernetes_structure_data(snapshot_source):
    """
    get_kubernetes_structure_data going to get structure data from connector 
    file which specified in snapshot as source field.
    Return structure data as json dictionary
    """
    kubernetes_structure_data = {}
    if json_source():
        qry = {'name': snapshot_source}
        dbname = config_value('MONGODB', 'dbname')
        sort = [sort_field('timestamp', False)]
        collection =config_value(DATABASE, collectiontypes[STRUCTURE])
        structure_docs = get_documents(collection=collection , dbname=dbname, sort= sort, query=qry, limit=1)
        logger.info('%s fetched %s number of documents: %s', Snapshot.LOGPREFIX, STRUCTURE, len(structure_docs))
        if structure_docs and len(structure_docs):
            kubernetes_structure_data = structure_docs[0]['json']
    else:
        kubernetes_structure_path = get_kubernetes_structure_path(snapshot_source)
        kubernetes_structure_data = json_from_file(kubernetes_structure_path)

    return kubernetes_structure_data

def make_kubernetes_snapshot_template(snapshot,node,kubernetes_snapshot_data):
    """
    make_kubernetes_snapshot_template prepare the data for db records and add data which got from 
    kubernetes cluster.
    This function is required  if we need to add kubernetes data to mongo db 
    """
    node_db_record_data= node_db_record(snapshot,node)
    node_db_record_data['json']=kubernetes_snapshot_data

    return node_db_record_data



def create_kube_apiserver_instance(snapshot,node):
    """
    create_kube_apiserver_instance creating kubernetes apiserver instance to have 
    communication and get response kubernetes apiserver 
    """
    snapshot_serviceAccount = get_field_value(snapshot,'serviceAccount')
    snapshot_namespace = get_field_value(snapshot,'namespace')
    node_type =  get_field_value(node,'type')
    snapshot_source = get_field_value(snapshot,'source')
    kubernetes_structure_data = get_kubernetes_structure_data(snapshot_source)
    api_instance = None
    service_account_secret = get_client_secret(kubernetes_structure_data,snapshot_serviceAccount,snapshot_namespace)

    cluster_url = get_field_value(kubernetes_structure_data,'clusterUrl')
    api_instance = create_kube_apiserver_instance_client(cluster_url,service_account_secret,node_type)
    return api_instance

def get_client_secret(kubernetes_structure_data,snapshot_serviceAccount,snapshot_namespace):
    """
    get_client_secret get service account from master snapshot and will 
    compare with other service accounts which allocated in kubernetes
    structure file and get secret of service account if it’s exist in
    structure file. Also check environment variables if service account
    secret isn’t exist in structure file with the name got from snap shot file.
    This function return secret as string which will use to get connection with 
    kubernetes cluster.

    """
    global Cache_namespace,Cache_secret
    if snapshot_namespace  == Cache_namespace:
        return Cache_secret


    namespaces = get_field_value(kubernetes_structure_data,'namespaces')
    service_account_secret = ""
    for namespace in namespaces :
        service_accounts = get_field_value(namespace,'serviceAccounts')
        for service_account in service_accounts :
            if snapshot_serviceAccount == service_account['name'] and namespace['namespace'] in snapshot_namespace :
                service_account_secret = get_field_value(service_account,'secret')
                if service_account_secret is not None:
                    Cache_secret= service_account_secret
                    Cache_namespace = snapshot_namespace
                    return service_account_secret
                else :
                    service_account_secret = get_vault_data(service_account['id'])
                    if  service_account_secret is not None:
                        Cache_secret= service_account_secret
                        Cache_namespace = snapshot_namespace
            

    
   
            

    
    if service_account_secret == "" :
        logger.error("\t\t ERROR : can not find secret for service account : %s" % (snapshot_serviceAccount))

    return service_account_secret 

def create_kube_apiserver_instance_client(cluster_url,service_account_secret,node_type):
    """ 
    Kubernetes library have several core kinds for getting data with cluster api server.
    For each Object type we have different core and we should use those function to get
    correct and reliable result. create_kube_apiserver_instance_client function pass 
    core instance due to node type.
    """
    node_type=node_type.lower()
    configuration = kubernetes.client.Configuration()
    token = '%s' % (service_account_secret)
    configuration.api_key={"authorization":"Bearer "+ token}
    configuration.host = cluster_url
    configuration.verify_ssl=False 
    configuration.debug = False
    client.Configuration.set_default(configuration)
    if node_type in ["pod","service","serviceaccount"]:
        api_client = client.CoreV1Api()
    if node_type in ["deployment","replicaset"]:
        api_client = client.AppsV1Api()
    if node_type in ["networkpolicy"]:
        api_client = client.NetworkingV1Api()
    if node_type in ["podsecuritypolicy"]:
        api_client = client.PolicyV1beta1Api()
    if node_type in ["rolebinding","role","clusterrole","clusterrolebinding"]:
        api_client = client.RbacAuthorizationV1beta1Api()
    return api_client

def get_kubernetes_snapshot_data(snapshot,node):
    """
    get_kubernetes_snapshot_data get api instance and due to snapshot and
    node and will request the data from kubernetes cluster.
    Snapshot file should have path in nodes so get_kubernetes_snapshot_data
    can use it here to find out which kind of object in which name space we need to get. 
    Also this function can use path to request the data from kubernetes apiserver.
    Some of the objects does not have any namespace so this function should 
    be able find out the  name space from path.
    In this function also kubernetes library exception are handled to find if
    any error happened.
    """
    path=  get_field_value(node,'paths')[0]
    node_type = get_field_value(node,'type')
    api_response = None
    object_name = path.split("/")[-1]
    snapshot_namespace = path.split("/")[3]
    api_instance = create_kube_apiserver_instance(snapshot,node)
    try:
        if node_type == "pod":
            snapshot_namespace = path.split("/")[3]
            api_response = api_instance.read_namespaced_pod(name=object_name,namespace=snapshot_namespace)

        if node_type == "deployment":
            snapshot_namespace = path.split("/")[4]
            api_response = api_instance.read_namespaced_deployment(name=object_name,namespace=snapshot_namespace)

        if node_type == "replicaset":
            snapshot_namespace = path.split("/")[4]
            api_response = api_instance.read_namespaced_replica_set(name=object_name,namespace=snapshot_namespace)

        if node_type == "service":
            snapshot_namespace = path.split("/")[3]
            api_response = api_instance.read_namespaced_service(name=object_name,namespace=snapshot_namespace)

        if node_type == "networkpolicy":
            snapshot_namespace = path.split("/")[4]
            api_response = api_instance.read_namespaced_network_policy(name=object_name,namespace=snapshot_namespace)

        if node_type == "podsecuritypolicy":
            api_response = api_instance.read_pod_security_policy(name=object_name)

        if node_type == "rolebinding":
            snapshot_namespace = path.split("/")[4]
            api_response = api_instance.read_namespaced_role_binding(name=object_name,namespace=snapshot_namespace)

        if node_type == "role":
            snapshot_namespace = path.split("/")[4]
            api_response = api_instance.read_namespaced_role(name=object_name,namespace=snapshot_namespace)
        
        if node_type == "clusterrolebinding":
            api_response = api_instance.read_cluster_role_binding(name=object_name)
        
        if node_type == "clusterrole":
            api_response = api_instance.read_cluster_role(name=object_name)
        
        if node_type == "serviceaccount":
            snapshot_namespace = path.split("/")[3]
            api_response = api_instance.read_namespaced_service_account(name=object_name,namespace=snapshot_namespace)

    except Exception as ex :
        logger.info('\t\tERROR : error in calling api for getting information %s : %s',node_type, object_name)
        logger.info('\t\tERROR : %s',ex)
        print(traceback.format_exc())

        
    api_response_dict = todict(api_response)  
    return api_response_dict

def todict(obj):
    """
    todict function convert data to serialized json.
    """
    if hasattr(obj, 'attribute_map'):
        result = {}
        for k,v in getattr(obj, 'attribute_map').items():
            val = getattr(obj, k)
            if val is not None:
                result[v] = todict(val)
        return result
    elif type(obj) == list:
        return [todict(x) for x in obj]
    elif type(obj) == datetime:
        return str(obj)
    else:
        return obj

def node_db_record(snapshot,node): 
    """
    node_db_record add additional fields to prepare the data for inserting 
    in database.
    """
    collection = node['collection'] if 'collection' in node else COLLECTION
    data = {
    "structure":"kubernetes",
    "reference": snapshot['namespace'],
    "contentType": "json",
    "source": snapshot['source'],
    "paths": get_field_value(node,"paths"),
    "timestamp": int(time.time() * 1000),
    "queryuser": snapshot['serviceAccount'],
    "checksum":hashlib.md5("{}".encode('utf-8')).hexdigest(),
    "node": node,
    "snapshotId":node['snapshotId'],
    "collection": collection.replace('.', '').lower(),
    "json": {}
    }

    return data

def get_lits(snapshot,node):
    node_type = get_field_value(node,'type')
    master_snapshot_func = {
        'pod' : get_list_namespaced_pods,
        'networkpolicy' : get_list_namespaced_network_policy,
        'podsecuritypolicy' : get_list_namespaced_pod_security_policy,
        'role':get_list_namespaced_role,
        'rolebinding' : get_list_namespaced_role_binding,
        'clusterrole':get_list_cluster_role,
        'clusterrolebinding' : get_list_cluster_role_binding,
        'serviceaccount' : get_list_namespaced_service_account,
    }
    list_item=[]
    try:
         return master_snapshot_func[node_type](snapshot,node)

    except Exception as ex :
        logger.info('\t\tERROR : error in calling api for getting information %s ',node_type)
        logger.info('\t\tERROR : %s',ex)
        print(traceback.format_exc())

    return list_item

def get_list_namespaced_pods(snapshot,node):

    snapshot_namespaces = get_field_value(snapshot,'namespace')
    pod_items = []
    api_instance = create_kube_apiserver_instance(snapshot,node)
    for snapshot_namespace in snapshot_namespaces:
        api_response = api_instance.list_namespaced_pod(namespace=snapshot_namespace)
        api_response_dict = todict(api_response) 
        api_response_dict_items = get_field_value(api_response_dict,'items')
        for api_response_dict_item in api_response_dict_items :
            pod_name = get_field_value(api_response_dict_item,'metadata.name')
            pod_path = "api/v1/namespaces/%s/pods/%s" % (snapshot_namespace,pod_name)
            pod_items.append({
                'namespace': snapshot_namespace,
                'paths':[
                    pod_path
                ]
            })
    return pod_items

def get_list_namespaced_network_policy(snapshot,node):
    snapshot_namespaces = get_field_value(snapshot,'namespace')
    network_policy_items = []
    api_instance = create_kube_apiserver_instance(snapshot,node)
    for snapshot_namespace in snapshot_namespaces:
        api_response = api_instance.list_namespaced_network_policy(namespace=snapshot_namespace)
        api_response_dict = todict(api_response) 
        api_response_dict_items = get_field_value(api_response_dict,'items')
        for api_response_dict_item in api_response_dict_items :
            network_policy_name = get_field_value(api_response_dict_item,'metadata.name')
            network_policy_path = "apis/networking.k8s.io/v1/namespaces/%s/networkpolicies/%s" % (snapshot_namespace,network_policy_name)
            network_policy_items.append({
                'namespace': snapshot_namespace,
                'paths':[
                    network_policy_path
                ]
            })
    return network_policy_items

def get_list_namespaced_pod_security_policy(snapshot,node):
    snapshot_namespaces = get_field_value(snapshot,'namespace')
    pod_security_policy_items = []
    api_instance = create_kube_apiserver_instance(snapshot,node)
    for snapshot_namespace in snapshot_namespaces:
        api_response = api_instance.list_pod_security_policy()
        api_response_dict = todict(api_response) 
        api_response_dict_items = get_field_value(api_response_dict,'items')
        for api_response_dict_item in api_response_dict_items :
            pod_security_policy_name = get_field_value(api_response_dict_item,'metadata.name')
            pod_security_policy_path = "apis/policy/v1beta1/podsecuritypolicies/%s" % (pod_security_policy_name)
            pod_security_policy_items.append({
                'namespace': snapshot_namespace,
                'paths':[
                    pod_security_policy_path
                ]
            })
    return pod_security_policy_items

def get_list_namespaced_role(snapshot,node):
    snapshot_namespaces = get_field_value(snapshot,'namespace')
    role_items = []
    api_instance = create_kube_apiserver_instance(snapshot,node)
    for snapshot_namespace in snapshot_namespaces:
        api_response = api_instance.list_namespaced_role(namespace=snapshot_namespace)
        api_response_dict = todict(api_response) 
        api_response_dict_items = get_field_value(api_response_dict,'items')
        for api_response_dict_item in api_response_dict_items :
            role_binding_name = get_field_value(api_response_dict_item,'metadata.name')
            role_binding_path = "apis/rbac.authorization.k8s.io/v1beta1/namespaces/%s/roles/%s" % (snapshot_namespace,role_binding_name)
            role_items.append({
                'namespace': snapshot_namespace,
                'paths':[
                    role_binding_path
                ]
            })
    return role_items


def get_list_namespaced_role_binding(snapshot,node):
    snapshot_namespaces = get_field_value(snapshot,'namespace')
    role_binding_items = []
    api_instance = create_kube_apiserver_instance(snapshot,node)
    for snapshot_namespace in snapshot_namespaces:
        api_response = api_instance.list_namespaced_role_binding(namespace=snapshot_namespace)
        api_response_dict = todict(api_response) 
        api_response_dict_items = get_field_value(api_response_dict,'items')
        for api_response_dict_item in api_response_dict_items :
            role_binding_name = get_field_value(api_response_dict_item,'metadata.name')
            role_binding_path = "apis/rbac.authorization.k8s.io/v1beta1/namespaces/%s/rolebindings/%s" % (snapshot_namespace,role_binding_name)
            role_binding_items.append({
                'namespace': snapshot_namespace,
                'paths':[
                    role_binding_path
                ]
            })
    return role_binding_items

def get_list_cluster_role(snapshot,node):
    snapshot_namespaces = get_field_value(snapshot,'namespace')
    cluster_role_items = []
    api_instance = create_kube_apiserver_instance(snapshot,node)
    for snapshot_namespace in snapshot_namespaces:
        api_response = api_instance.list_cluster_role()
        api_response_dict = todict(api_response) 
        api_response_dict_items = get_field_value(api_response_dict,'items')
        for api_response_dict_item in api_response_dict_items :
            cluster_role_name = get_field_value(api_response_dict_item,'metadata.name')
            cluster_role_path = "apis/rbac.authorization.k8s.io/v1beta1/clusterroles/%s" % (cluster_role_name)
            cluster_role_items.append({
                'namespace': snapshot_namespace,
                'paths':[
                    cluster_role_path
                ]
            })
    return cluster_role_items

def get_list_cluster_role_binding(snapshot,node):
    snapshot_namespaces = get_field_value(snapshot,'namespace')
    cluster_role_binding_items = []
    api_instance = create_kube_apiserver_instance(snapshot,node)
    for snapshot_namespace in snapshot_namespaces:
        api_response = api_instance.list_cluster_role()
        api_response_dict = todict(api_response) 
        api_response_dict_items = get_field_value(api_response_dict,'items')
        for api_response_dict_item in api_response_dict_items :
            cluster_role_binding_name = get_field_value(api_response_dict_item,'metadata.name')
            cluster_role_binding_path = "apis/rbac.authorization.k8s.io/v1beta1/clusterrolebindings/%s" % (cluster_role_binding_name)
            cluster_role_binding_items.append({
                'namespace': snapshot_namespace,
                'paths':[
                    cluster_role_binding_path
                ]
            })
    return cluster_role_binding_items

def get_list_namespaced_service_account(snapshot,node):
    snapshot_namespaces = get_field_value(snapshot,'namespace')
    service_account_items = []
    api_instance = create_kube_apiserver_instance(snapshot,node)
    for snapshot_namespace in snapshot_namespaces:
        api_response = api_instance.list_namespaced_service_account(namespace=snapshot_namespace)
        api_response_dict = todict(api_response) 
        api_response_dict_items = get_field_value(api_response_dict,'items')
        for api_response_dict_item in api_response_dict_items :
            service_account_name = get_field_value(api_response_dict_item,'metadata.name')
            service_account_path = "api/v1/namespaces/%s/serviceaccounts/%s" % (snapshot_namespace,service_account_name)
            service_account_items.append({
                'namespace': snapshot_namespace,
                'paths':[
                    service_account_path
                ]
            })
    return service_account_items

def generate_crawler_snapshot(snapshot,node,snapshot_data):
    node_type= get_field_value(node,'type')
    node_type = get_field_value_with_default(node, 'type',"")
    resource_items = []
    resource_items=get_lits(snapshot=snapshot,node=node)

    includeSnapshotConfig = get_from_currentdata("INCLUDESNAPSHOTS")
    includeSnapshots = get_from_currentdata("SNAPHSHOTIDS")
    ignoreExclusion = False
    ignoreNode = False
    if includeSnapshotConfig:
        if node['masterSnapshotId'] in includeSnapshots:
            ignoreExclusion = True
            ignoreNode = False
        else:
            ignoreNode  = True

    exclusions = get_from_currentdata(EXCLUSION).get('exclusions', [])
    resourceExclusions = {}
    if not ignoreExclusion:
        for exclusion in exclusions:
            if 'exclusionType' in exclusion and exclusion['exclusionType'] and exclusion['exclusionType'] == 'resource':
                if 'paths' in exclusion and isinstance(exclusion['paths'], list):
                    resourceExclusions[tuple(exclusion['paths'])] = exclusion

    snapshot_data[node['masterSnapshotId']] = []
    for index,resource_namespaced_item in enumerate(resource_items) :
        if isinstance(resource_namespaced_item['paths'], str):
            key = tuple([resource_namespaced_item['paths']])
        elif isinstance(resource_namespaced_item['paths'], list):
            key = tuple(resource_namespaced_item['paths'])
        else:
            key = None
        if key and key in resourceExclusions:
            logger.warning("Excluded from resource exclusions: %s", resource_namespaced_item['paths'])
            continue
        if not ignoreNode:
            snapshot_data[node['masterSnapshotId']].append({
                            "masterSnapshotId" : [node['masterSnapshotId']],
                            "snapshotId": '%s%s_%s' % (node['masterSnapshotId'],resource_namespaced_item['namespace'], str(index)),
                            "type": node_type,
                            "collection": node['collection'],
                            "paths": resource_namespaced_item['paths'],
                            "status" : "active",
                            "validate" : node['validate'] if 'validate' in node else True
                        })
    return snapshot_data



def populate_kubernetes_snapshot(snapshot, container=None):
    snapshot_nodes = get_field_value(snapshot,'nodes')
    snapshot_data, valid_snapshotids = validate_snapshot_nodes(snapshot_nodes)
    dbname = config_value('MONGODB', 'dbname')
    if valid_snapshotids  and snapshot_nodes:
        logger.debug(valid_snapshotids)
        try :
            for node in snapshot_nodes:
                validate = node['validate'] if 'validate' in node else True
                logger.info(node)
                if 'snapshotId' in node:
                    if validate:
                        kubernetes_snapshot_data = get_kubernetes_snapshot_data(snapshot,node) 
                        if kubernetes_snapshot_data :
                            error_str = kubernetes_snapshot_data.pop('error', None)
                            kubernetes_snapshot_template = make_kubernetes_snapshot_template(
                                snapshot,
                                node,
                                kubernetes_snapshot_data
                            )
                            if get_dbtests():
                                if get_collection_size(kubernetes_snapshot_template['collection']) == 0:
                                    #Creating indexes for collection
                                    create_indexes(
                                        kubernetes_snapshot_template['collection'], 
                                        config_value(DATABASE, DBNAME), 
                                        [
                                            ('snapshotId', pymongo.ASCENDING),
                                            ('timestamp', pymongo.DESCENDING)
                                        ]
                                    )
                                    create_indexes(
                                        kubernetes_snapshot_template['collection'], 
                                        config_value(DATABASE, DBNAME), 
                                        [
                                            ('_id', pymongo.DESCENDING),
                                            ('timestamp', pymongo.DESCENDING),
                                            ('snapshotId', pymongo.ASCENDING)
                                        ]
                                    )
                                insert_one_document(kubernetes_snapshot_template, kubernetes_snapshot_template['collection'], dbname,check_keys=False)
                            
                            snapshot_dir = make_snapshots_dir(container)
                            if snapshot_dir:
                                store_snapshot(snapshot_dir, kubernetes_snapshot_template)                                
                            if "masterSnapshotId" in node :
                                snapshot_data[node['snapshotId']] = node['masterSnapshotId']
                            elif "snapshotId" in node :
                                snapshot_data[node['snapshotId']] = False if error_str else True
                        else:
                            node['status'] = 'inactive'
                elif 'masterSnapshotId' in node:
                    snapshot_data = generate_crawler_snapshot(snapshot,node,snapshot_data)
                


        except Exception as ex:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                logger.error('can not connect to kubernetes cluster: %s ', ex )
                logger.error('\t ERROR INFO : \n \tfile name : %s\n \tline : %s\n \ttype : %s\n \tobject : %s',fname,exc_tb.tb_lineno,exc_type,exc_obj)
                print(traceback.format_exc())
                raise ex
    return snapshot_data



"""
Define an interface for creating an object, but let subclasses decide
which class to instantiate. Factory Method lets a class defer
instantiation to subclasses.
"""
import time
import sys
import json
import logging
import os
import re
import pymongo
import subprocess
from processor.helper.json.json_utils import get_field_value, json_from_file, save_json_to_file
from processor.helper.config.config_utils import get_test_json_dir, parsebool, config_value, generateid
from processor.helper.file.file_utils import exists_file, exists_dir, remove_file
from processor.database.database import COLLECTION, get_documents
from processor.comparison.comparison_functions import equality,\
    less_than, less_than_equal, greater_than, greater_than_equal, exists
from antlr4 import InputStream
from antlr4 import CommonTokenStream
from antlr4.error.ErrorListener import ErrorListener, ConsoleErrorListener
from processor.comparison.comparisonantlr.comparatorLexer import comparatorLexer
from processor.comparison.comparisonantlr.comparatorParser import comparatorParser
from processor.comparison.comparisonantlr.rule_interpreter import RuleInterpreter
from processor.helper.config.rundata_utils import get_dbtests, get_from_currentdata
from processor.logging.log_handler import getlogger
import traceback


logger = getlogger()
COMPARATOR_V0_1 = '0_1'
COMPARATOR_V0_2 = '0_2'
OPERATORS = {
    'eq': equality,
    'neq': equality,
    'lt': less_than,
    'le': less_than_equal,
    'gt': greater_than,
    'ge': greater_than_equal,
    'not': None,
    'exist': exists
}
MATHOPERATORS = ['lt', 'le', 'gt', 'ge', 'eq', 'neq']
TESTCASEV1 = 1
TESTCASEV2 = 2

class MyConsoleErrorListener(ErrorListener):

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        logger.debug("******line " + str(line) + ":" + str(column) + " " + msg)

ConsoleErrorListener.INSTANCE = MyConsoleErrorListener()


def version_str(version):
    """Convert numeric version to string eg: 0.1 => 0_1"""
    value = version.replace('.', '_') if version else version
    return value


def get_operator_roperand(value):
    roperand = None
    op = 'exist'
    is_not = False
    extras = None
    if value:
        if value.startswith('not') or value.startswith('neq'):
            is_not = True
            value = value.replace('not', '') if value.startswith('not') \
                else value.replace('neq', 'eq')
            value = value.strip()
        vals = value.split()
        if vals and vals[0] in OPERATORS:
            op = vals[0]
            if len(vals) > 1:
                roperand = ' '.join(vals[1:])
        else:
            roperand = value
        if roperand:
            is_math = True if op in MATHOPERATORS else False
            extras, roperand = adapt_roperand(roperand, is_math)
    return is_not, op, roperand, extras


def adapt_roperand(roperand, is_math=False):
    value = roperand
    extras = None
    if is_math:
        try:
            value = int(roperand)
        except:
            pass
    if value and not isinstance(value, int):
        if value[0] == '"' and value[-1] == '"':
            value = value.replace('"', '')
        elif value[0] == "'" and value[-1] == "'":
            value = value.replace("'", "")
        extended_value = interpret_additional_operations(value)
        if extended_value:
            value = list(extended_value.values()).pop(0)
            extras = list(extended_value.keys())
    return extras, value


def match_array_len(_, roperand):
    value = int(roperand[4:-1])
    return {'len': value}


def fetch_snapshot_attribute(m, roperand):
    pass


def interpret_additional_operations(roperand):
    value = None
    actions = ((r'len\(\d+\)', match_array_len),
               (r'\{2\}\.(.*)', fetch_snapshot_attribute))
    # try:
    for regex, action in actions:
        m = re.match(regex, roperand, re.I)
        if m:
            logger.debug(m.groups(), regex)
            value = action(m, roperand)
            break
    # except:
    #     pass
    return value


def opa_binary():
    opa_exe = None
    opa_enabled = parsebool(config_value("OPA", "opa"), False)
    if opa_enabled:
        opa_exe = os.getenv('OPAEXE', None)
        # print(opa_exe)
        if opa_exe and exists_file(opa_exe):
            # print('%' * 50)
            pass
        else:
            # print('$' * 50)
            opa_exe = config_value("OPA", "opaexe")
            if opa_exe and exists_file(opa_exe):
                pass
            else:
                opa_exe = None
    if not opa_exe:
        try:
            subprocess.Popen(['opa', "version"], stdout=subprocess.DEVNULL)
            opa_exe = "opa"
        except FileNotFoundError:
            opa_exe = None

    return opa_exe

def get_rego_rule_filename(rego_file, container):
    rego_file_name = None
    json_dir = get_test_json_dir()
    if exists_dir(json_dir):
        rego_file_name = '%s/%s/%s' % (json_dir, container, rego_file)
        if exists_file(rego_file_name):
            pass
        else:
            rego_file_name = None
    return rego_file_name


def compare(inputjson, rulestr):
   pass


def import_from(module, name):
    try:
        module = __import__(module, fromlist=[name])
        return getattr(module, name)
    except:
        return

def import_module(module):
    try:
        module = __import__(module)
        return module
    except:
        logger.debug(traceback.format_exc())
        return False

class Comparator:
    """
    Declare the factory method, which returns a comparator object.
    Comparator also defines the default implementation of the factory
    method that returns a Concrete Comparator object based on the version.
    Call the factory method to create the comparator object.
    """

    def __init__(self, version, container, dbname, collection_data, testcase, excludedTestIds, includeTests):
        self.comparator = self._factory_method(version, container, dbname, collection_data,
                                               testcase, excludedTestIds, includeTests)

    @staticmethod
    def _factory_method(version, container, dbname, collection_data, testcase, excludedTestIds, includeTests):
        version_val = version_str(version)
        if version_val == COMPARATOR_V0_1:
            return ComparatorV01(container, dbname, collection_data, testcase, excludedTestIds, includeTests)
        elif version_val == COMPARATOR_V0_2:
            return ComparatorV02(container, dbname, collection_data, testcase, excludedTestIds, includeTests)
        else:
            return ComparatorV01(container, dbname, collection_data, testcase, excludedTestIds, includeTests)

    def validate(self):
        return self.comparator.validate()


class ComparatorV01:
    """Override the validate method to return to run comparator"""

    def __init__(self, container, dbname, collection_data, testcase, excludedTestIds, includeTests):
        self.container = container
        self.dbname = dbname
        self.collection_data = collection_data
        self.testcase = testcase
        self.excludedTestIds = excludedTestIds
        self.includeTests = includeTests
        loperand = get_field_value(testcase, 'attribute')
        value = get_field_value(testcase, 'comparison')
        rule = get_field_value(testcase, 'rule')
        isrego = get_field_value(testcase, 'type')
        self.snapshots = []
        if isrego and isrego == 'rego' or isrego == "python":
            self.format = TESTCASEV2
            self.rule = rule
            self.type = isrego
        elif loperand and value:
            self.format = TESTCASEV1
            self.snapshot_id = get_field_value(testcase, 'snapshotId')
            coll_val = get_field_value(self.collection_data, self.snapshot_id)
            self.collection = coll_val if coll_val else COLLECTION
            self.loperand = loperand
            self.is_not, self.op, self.roperand, self.extras = get_operator_roperand(value)
            self.type = 'prancer'
        elif rule:
            self.format = TESTCASEV2
            self.rule = rule
            self.type = 'prancer'
        else:
            self.format = None

    def log_compliance_info(self, test_id):
        logger.critical('\tTESTID: %s', test_id)
        logger.critical('\t\tSNAPSHOTID: %s', self.testcase['snapshotId'][0])
        if self.snapshots:
            snapshot = self.snapshots[0]
            logger.critical('\t\tPATHS: ')
            for path in snapshot.get('paths', []):
                logger.critical('\t\t\t %s', path)

    def process_rego_test_case(self):
        tid = '%d_%s' % (int(time.time() * 1000000), generateid(None))
        results = []
        inputjson = {}
        result = False
        rule_expr = get_field_value(self.testcase, 'eval')
        if not rule_expr:
            rule_expr = get_field_value(self.testcase, 'evals')
            if rule_expr:
                del self.testcase['evals']
        if not rule_expr:
            rule_expr = 'data.rule.rulepass'
        testId = 'MISSING ID'
        isMasterTest = False
        if 'testId' in self.testcase:
            testId = self.testcase['testId']
        elif 'masterTestId' in self.testcase:
            testId = self.testcase['masterTestId']
            isMasterTest = True
        sid = self.testcase['snapshotId'][0]
        toExclude, snapshot_doc = self.get_snaphotid_doc(sid, testId, isMasterTest)
        if toExclude:
            logger.warn('\t\tWARN: Excluded test case: %s' % testId)
            logger.warn('\t\tRESULT: SKIPPED')
            # msg = 'Excluded testcase because of testId: %s' % testId
            # results.append({'eval': 'data.rule.rulepass', 'result': 'skipped', 'message': msg})
            return results

                            
        # logger.critical('\t\tEVAL: %s', rule_expr)
        opa_exe = opa_binary()        
        if not opa_exe:
            # print('*' * 50)
            logger.error('\t\tERROR: OPA binary not found!')
            logger.error('\t\tRESULT: FAILED')
            results.append({'eval': 'data.rule.rulepass', 'result': "passed" if result else "failed", 'message': ''})
            return results
        if len(self.testcase['snapshotId'])==1:
            inputjson = snapshot_doc
            if inputjson is None:
                logger.info('\t\tERROR: Missing snapshot')
        else:
            ms_id = dict(zip(self.testcase['snapshotId'], self.testcase['masterSnapshotId']))
            for sid in self.testcase['snapshotId']:
                toExclude, snapshot_doc = self.get_snaphotid_doc(sid, testId, isMasterTest)
                if toExclude:
                    logger.error('Excluded testcase because of testId: %s' % testId)
                inputjson.update({ms_id[sid]: snapshot_doc})
        results = []
        if inputjson:
            save_json_to_file(inputjson, '/tmp/input_%s.json' % tid)
            rego_rule = self.rule
            rego_match=re.match(r'^file\((.*)\)$', rego_rule, re.I)
            if rego_match:
                rego_file = self.rego_rule_filename(rego_match.groups()[0], self.container)
                if rego_file:
                    pass
                else:
                    rego_file = None
            else:
                rego_txt = [
                    "package rule",
                    "default rulepass = false",
                    "rulepass = true{",
                    "   %s" % rego_rule,
                    "}", ""
                ]
                rego_file = '/tmp/input_%s.rego' % tid
                open(rego_file, 'w').write('\n'.join(rego_txt))
            if rego_file:
                if isinstance(rule_expr, list):
                    result = os.system('%s eval -i /tmp/input_%s.json -d %s "data.rule" > /tmp/a_%s.json' % (opa_exe, tid, rego_file, tid))
                    if result != 0 :
                        self.log_compliance_info(testId)
                        logger.error("\t\tERROR: have problem in running opa binary")
                        self.log_rego_error(json_from_file("/tmp/a_%s.json" % tid, object_pairs_hook=None))
                else:
                    result = os.system('%s eval -i /tmp/input_%s.json -d %s "%s" > /tmp/a_%s.json' % (opa_exe, tid, rego_file, rule_expr, tid))
                    if result != 0 :
                        self.log_compliance_info(testId)
                        logger.error("\t\tERROR: have problem in running opa binary")
                        self.log_rego_error(json_from_file("/tmp/a_%s.json" % tid, object_pairs_hook=None))

                resultval = json_from_file('/tmp/a_%s.json' % tid)
                if resultval and "errors" in resultval and resultval["errors"]:
                    if isinstance(rule_expr, list):
                        if rule_expr[0] and "eval" in rule_expr[0]:
                            results.append({'eval': rule_expr[0].get("eval"), 'result': "failed", 'message': ''})
                        else:
                            results.append({'eval': "data.rule", 'result': "failed", 'message': ''})
                    else:
                        results.append({'eval': rule_expr, 'result': "failed", 'message': ''})
                    self.log_compliance_info(testId)
                    logger.critical('\t\tTITLE: %s', self.testcase.get('title', ""))
                    logger.critical('\t\tDESCRIPTION: %s', self.testcase.get('description', ""))
                    logger.critical('\t\tRULE: %s', self.testcase.get('rule', ""))
                    logger.critical('\t\tERROR: %s', str(resultval["errors"]))
                    logger.critical('\t\tREMEDIATION: %s', self.testcase.get('remediation_description', ""))
                    logger.error('\t\tRESULT: FAILED')
                    # logger.error(str(resultval["errors"]))
                elif resultval:
                    if isinstance(rule_expr, list):
                        resultdict = resultval['result'][0]['expressions'][0]['value']
                        for val in rule_expr:
                            if 'eval' in val: 
                                evalfield = val['eval'].rsplit('.', 1)[-1]
                                evalmessage = val['message'].rsplit('.', 1)[-1] if "message" in val else ""
                                if evalfield in resultdict:
                                    if isinstance(resultdict[evalfield], bool):
                                        result = parsebool(resultdict[evalfield])
                                    else:
                                        if logger.level == logging.DEBUG:
                                            self.log_compliance_info(testId)
                                            logger.warning('\t\tRESULT: SKIPPED')
                                        continue
                                elif evalmessage in resultdict:
                                    result = False
                                else:
                                    if logger.level == logging.DEBUG:
                                        self.log_compliance_info(testId)
                                        logger.warning('\t\tRESULT: SKIPPED')
                                    continue
                                msg = resultdict[evalmessage] if not result and evalmessage in resultdict else ""
                                results.append({
                                    'eval': val["eval"], 
                                    'result': "passed" if result else "failed", 
                                    'message': msg,
                                    'id' : val.get("id"),
                                    'remediation_description' : val.get("remediationDescription"),
                                    'remediation_function' : val.get("remediationFunction"),
                                })
                                # logger.info('\t\tERROR: %s', resultval)
                                self.log_compliance_info(testId)
                                self.log_result(results[-1])
                    else:
                        resultbool = resultval['result'][0]['expressions'][0]['value'] # [get_field_value(resultval, 'result[0].expressions[0].value')
                        result = parsebool(resultbool)
                        results.append({'eval': rule_expr, 'result': "passed" if result else "failed", 'message': ''})
                        # logger.info('\t\tERROR: %s', resultval)
                        self.log_compliance_info(testId)
                        self.log_result(results[-1])
                        if results[-1]['result'] == 'failed':
                            logger.error('\t\tERROR: %s', json.dumps(dict(resultval)))

            else:
                if logger.level == logging.DEBUG:
                    self.log_compliance_info(testId)
                    logger.info('\t\tERROR: %s missing', rego_match.groups()[0])
                    logger.warning('\t\tRESULT: SKIPPED')
                # results.append({'eval': rule_expr, 'result': "passed" if result else "failed", 'message': ''})
                # self.log_result(results[-1])
            remove_file('/tmp/input_%s.json' % tid)
            remove_file('/tmp/a_%s.json' % tid)
        else:
            results.append({'eval': rule_expr, 'result': "passed" if result else "failed", 'message': ''})
            self.log_result(results[-1])
        return results


    def process_python_test_case(self) -> list:
        # setting path of tmp to allow import from temp
        sys.path.append('/tmp/')
        results = []
        inputjson = {}
        result = False

        # getting rule_expr value from evals or eval
        rule_expr = get_field_value(self.testcase, 'eval')
        if not rule_expr:
            rule_expr = get_field_value(self.testcase, 'evals')
            if rule_expr:
                del self.testcase['evals']
        if not rule_expr:
            rule_expr = 'data.rule.rulepass'

        # getting testId or MasterTestId
        testId = 'MISSING ID'
        isMasterTest = False
        if 'testId' in self.testcase:
            testId = self.testcase['testId']
        elif 'masterTestId' in self.testcase:
            testId = self.testcase['masterTestId']
            isMasterTest = True
        

        snapshot_id = self.testcase['snapshotId'][0]

        # getting snapshot and exclude status
        toExclude, snapshot_doc = self.get_snaphotid_doc(snapshot_id, testId, isMasterTest)
        if toExclude:
            logger.warn('\t\tWARN: Excluded test case: %s' % testId)
            logger.warn('\t\tRESULT: SKIPPED')
            return results
        
        # python file name
        python_testcase = self.rule.split(".")[0]
        # python function name

        if len(self.testcase['snapshotId']) == 1:
            inputjson = snapshot_doc
            if inputjson is None:
                logger.info('\t\tERROR: Missing snapshot')
        else:
            ms_id = dict(zip(self.testcase['snapshotId'], self.testcase['masterSnapshotId']))
            for sid in self.testcase['snapshotId']:
                toExclude, snapshot_doc = self.get_snaphotid_doc(sid, testId, isMasterTest)
                if toExclude:
                    logger.error('Excluded testcase because of testId: %s' % testId)
                inputjson.update({ms_id[sid]: snapshot_doc})
        results = []

        test_rule = self.rule
        rule_matched = re.match(r'^file\((.*)\)$', test_rule, re.I)
        if rule_matched:
            rego_file_name = self.rego_rule_filename(rule_matched.groups()[0], self.container)
            if not rego_file_name:
                python_testcase = "processor.comparison.rules.%s.%s"%(self.snapshots[0]["type"],rule_matched.groups()[0].split(".")[0])
                module = import_module(python_testcase)
                if not module and logger.level == logging.DEBUG:
                    self.log_compliance_info(testId)
                    logger.error('\t\tERROR: %s missing', rule_matched.groups()[0])
                    logger.warning('\t\tRESULT: SKIPPED')
                    return results
            else:            
                python_testcase = rule_matched.groups()[0].split(".")[0]
        
        if isinstance(rule_expr, list):
            for rule in rule_expr:
                function_name = rule["eval"].rsplit(".", 1)[-1] if "eval" in rule else ""
                evalmessage = rule['message'].rsplit('.', 1)[-1] if "message" in rule else ""

                test_function = import_from(python_testcase, function_name)
                if not test_function:
                    self.log_compliance_info(testId)
                    logger.info('\t\tERROR: %s missing', rule_matched.groups()[0])
                    logger.warning('\t\tRESULT: SKIPPED')
                    return results
                
                paths = self.snapshots[0]["paths"]

                result = test_function(inputjson, kwargs={"paths": paths})

                if result.get("issue") == True:
                    result["result"] = "failed"
                    
                    self.log_compliance_info(testId)
                    self.log_result(result)
                    json_result = {
                        'eval': rule["eval"], 
                        'result': result["result"], 
                        'message': result.get(evalmessage, ""),
                        'id' : rule.get("id"),
                        'remediation_description' : rule.get("remediationDescription"),
                        'remediation_function' : rule.get("remediationFunction"),
                    }

                    if result.get("errors"):
                        json_result["errors"] = result.get("errors",[])

                    results.append(json_result)

                elif result.get("issue") == False:
                    if logger.level == logging.DEBUG:
                        self.log_compliance_info(testId)
                        logger.info('\t\tERROR: %s missing', rule_matched.groups()[0])
                        logger.warning('\t\tRESULT: SKIPPED')
                
                elif result["issue"] == None:
                    logger.error("\t\tERROR: have problem in running test")
                    logger.error(result[evalmessage])

        return results
                

    def log_rego_error(self, result):
        """ 
        log error raised while running the rego file
        """
        if result and "errors" in result and result.get("errors"):
            error = result.get("errors")[0]
            location = error.get("location")
            logger.error('\t\t  %s' % error.get("message"))
            
            if location:
                file_location = "/".join(location.get("file", "").split("/")[3:])
                logger.error("\t\t  location : %s" % file_location)
                logger.error("\t\t  row : %s" % location.get("row"))
                logger.error("\t\t  col : %s" % location.get("col"))
                
    def log_result(self, result):
        if result.get("result") == "passed":
            logger.critical('\t\tTITLE: %s', self.testcase.get('title', ""))
            logger.critical('\t\tRULE: %s', self.testcase.get('rule', ""))
            logger.critical('\t\tRESULT: %s', result.get("result"), extra={"type" : "SUCCESS"})            
        else:
            logger.critical('\t\tTITLE: %s', self.testcase.get('title', ""))
            logger.critical('\t\tDESCRIPTION: %s', self.testcase.get('description', ""))
            logger.critical('\t\tRULE: %s', self.testcase.get('rule', ""))
            logger.critical('\t\tERROR: %s', result.get("message", ""))
            logger.critical('\t\tREMEDIATION: %s', result.get('remediation_description', ""))
            logger.error('\t\tRESULT: %s', result.get("result"))

    def get_snaphotid_doc_old(self, sid, container):
        doc = None
        json_dir = get_test_json_dir()
        if exists_dir(json_dir):
            fname = '%s/%s/snapshots/%s' % (json_dir, container, sid)
            if exists_file(fname):
                json_data = json_from_file(fname)
                if json_data and 'json' in json_data:
                    doc = json_data['json']
                    snapshot = {
                        'id': json_data['snapshotId'],
                        'structure': json_data['structure'],
                        'reference': json_data['reference'],
                        'source': json_data['source'],
                        'collection': json_data['collection'],
                        'type': json_data.get("node", {}).get('type'),
                        'region' : json_data.get('region', "")
                    }
                    if 'paths' in json_data:
                        snapshot['paths'] = json_data['paths']
                    else:
                        snapshot['path'] = json_data['path']

                    self.snapshots.append(snapshot)
        return doc

    def exclude_test_case(self, doc, testId, isMasterTest=False):
        toExclude = False
        if isMasterTest and testId:
            if testId in self.includeTests:
                toExclude = False
            elif testId in self.excludedTestIds:
                path = doc['paths'][0]
                toExclude = True if path in self.excludedTestIds[testId] else False
        return toExclude
        pass


    def get_snaphotid_doc(self, sid, testId, isMasterTest=False):
        tobeExcluded = False
        doc = None
        isdb_fetch = get_dbtests()
        if isdb_fetch:
            dbname = self.dbname
            coll = self.collection_data[sid] if sid in self.collection_data else COLLECTION
            docs = get_documents(coll, {'snapshotId': sid}, dbname,
                                 sort=[('timestamp', pymongo.DESCENDING)], limit=1)
            logger.debug('Number of Snapshot Documents: %s', len(docs))
            if docs and len(docs):
                tobeExcluded = self.exclude_test_case(docs[0], testId, isMasterTest)
                doc = docs[0]['json']
                snapshot = {
                    'id': docs[0]['snapshotId'],
                    'structure': docs[0]['structure'],
                    'reference': docs[0]['reference'],
                    'source': docs[0]['source'],
                    'collection': docs[0]['collection'],
                    'type': docs[0].get("node", {}).get('type'),
                    'region' : docs[0].get('region', ""),
                    'resourceTypes' : docs[0].get("node", {}).get('resourceTypes', []),
                }
                if 'paths' in docs[0]:
                    snapshot['paths'] = docs[0]['paths']
                else:
                    snapshot['path'] = docs[0]['path']
                self.snapshots.append(snapshot)
        else:
            json_dir = '%s%s' % (get_test_json_dir(), self.container)
            if exists_dir(json_dir):
                fname = '%s/snapshots/%s' % (json_dir, sid)
                if exists_file(fname):
                    json_data = json_from_file(fname)
                    if json_data and 'json' in json_data:
                        tobeExcluded = self.exclude_test_case(json_data, testId, isMasterTest)
                        doc = json_data['json']
                        snapshot_val = {
                            'id': json_data['snapshotId'],
                            'structure': json_data['structure'],
                            'reference': json_data['reference'],
                            'source': json_data['source'],
                            'collection': json_data['collection'],
                            'type': json_data.get("node", {}).get('type'),
                            'region' : json_data.get('region', ""),
                            'resourceTypes' : json_data.get("node", {}).get('resourceTypes', []),
                        }
                        if 'paths' in json_data:
                            snapshot_val['paths'] = json_data['paths']
                        else:
                            snapshot_val['path'] = json_data['path']

                        # singletest = get_from_currentdata(SINGLETEST)
                        # if singletest:
                        #     snapshot_val['json'] = doc
                        self.snapshots.append(snapshot_val)
        return tobeExcluded, doc


    def rego_rule_filename(self, rego_file, container):
        rego_file_name = None
        if 'dirpath' in self.testcase and self.testcase['dirpath']:
            rego_file_name = '%s/%s' % (self.testcase['dirpath'], rego_file)
            if exists_file(rego_file_name):
                pass
            else:
                rego_file_name = None
            return  rego_file_name
        isdb_fetch = get_dbtests()
        #It give same value for DB and SNAPSHOT, So for SNAPSHOT, we'll check it in 
        #db first and if file isn't there, then we are fetching it from file path '''
        
        if isdb_fetch:
            dbname = self.dbname
            coll = 'structures'
            docs = get_documents(coll, { 'type': 'others', "container" : container}, dbname,
                                 sort=[('timestamp', pymongo.DESCENDING)], limit=1)
            # print('Number of other Documents: %s' % len(docs))
            logger.debug('Number of other Documents: %s', len(docs))
            if docs and len(docs):
                doc = docs[0]['json']
                if doc and 'file' in doc and isinstance(doc['file'], list):
                    for file_doc in doc['file']:
                        name = get_field_value(file_doc, 'name')
                        # print(name, rego_file)
                        if name == rego_file:
                            content = get_field_value(file_doc, 'container_file')
                            if content:
                                rego_file_name = '/tmp/%s' % rego_file
                                open(rego_file_name, 'w', encoding="utf-8").write(content)
                                return rego_file_name
                # print(doc)

        json_dir = get_test_json_dir()
        if exists_dir(json_dir):
            rego_file_name = '%s/%s/%s' % (json_dir, container, rego_file)
            if exists_file(rego_file_name):
                pass
            else:
                rego_file_name = None
        return rego_file_name

    def get_connector_data(self):
        """ get connector data from snapshot """
        connector_data = {}
        if self.snapshots:
            isdb_fetch = get_dbtests()
            if isdb_fetch:
                connectors = get_documents(
                    "structures",
                    query={
                        "name" : self.snapshots[0].get("source"),
                        "type" : "structure",
                        "container": self.container
                    },
                    dbname=self.dbname,
                    limit=1
                )
                connector_data = connectors[0].get("json", {}) if connectors else {}
            else:
                json_test_dir = get_test_json_dir()
                snapshot_source = self.snapshots[0].get("source")
                file_name = '%s.json' % snapshot_source if snapshot_source and not \
                    snapshot_source.endswith('.json') else snapshot_source
                connector_path = '%s/../%s' % (json_test_dir, file_name)
                if exists_file(connector_path):
                    connector_data = json_from_file(connector_path)

        return connector_data

    def validate(self):
        result_val = [{"result": "failed"}]
        if self.format == TESTCASEV1:
            if self.snapshot_id:
                docs = get_documents(self.collection, dbname=self.dbname,
                                     sort=[('timestamp', pymongo.DESCENDING)],
                                     query={'snapshotId': self.snapshot_id},
                                     limit=1)
                logger.info('Number of Snapshot Documents: %s', len(docs))
                if docs and len(docs):
                    self.data = docs[0]['json']
                    if self.op in OPERATORS and OPERATORS[self.op]:
                        result = OPERATORS[self.op](self.data, self.loperand, self.roperand,
                                                    self.is_not, self.extras)
                        result_val[0]["result"] = "passed" if result else "failed"
                        result_val[0]["snapshots"] = [{
                            'id': docs[0]['snapshotId'],
                            'structure': docs[0]['structure'],
                            'reference': docs[0]['reference'],
                            'source': docs[0]['source'],
                            'collection': docs[0]['collection']
                        }]
                        if "paths" in docs[0]:
                            result_val[0]["snapshots"][0]["paths"] = docs[0]["paths"]
                        else:
                            result_val[0]["snapshots"][0]["path"] = docs[0]["path"]
                else:
                    result_val[0].update({
                        "result": "skipped",
                        "message": "Missing documents for the snapshot"
                    })
            else:
                result_val[0].update({
                    "result": "skipped",
                    "message": "Missing snapshotId for testcase"
                })
        elif self.format == TESTCASEV2:
            if self.type == 'rego':
                results = self.process_rego_test_case()
                result_val = []
                connector_data = self.get_connector_data()
                for result in results:
                    result['snapshots'] = self.snapshots
                    result['autoRemediate'] = connector_data.get("autoRemediate", False)
                    result_val.append(result)
            
            elif self.type == "python":
                results = self.process_python_test_case()
                result_val = []
                connector_data = self.get_connector_data()
                for result in results:
                    result['snapshots'] = self.snapshots
                    result['autoRemediate'] = connector_data.get("autoRemediate", False)
                    result_val.append(result)
            
            else:
                # logger.info('#' * 75)
                logger.critical('\tTESTID: %s', self.testcase['testId'])
                input_stream = InputStream(self.rule)
                lexer = comparatorLexer(input_stream)
                stream = CommonTokenStream(lexer)
                parser = comparatorParser(stream)
                tree = parser.expression()
                children = []
                for child in tree.getChildren():
                    children.append((child.getText()))
                # logger.info('*' * 50)
                logger.debug("All the parsed tokens: %s", children)
                otherdata = {'dbname': self.dbname, 'snapshots': self.collection_data, 'container': self.container}
                r_i = RuleInterpreter(children, **otherdata)
                # result = r_i.compare()
                l_val, r_val, result = r_i.compare()
                result_val[0]["result"] = "passed" if result else "failed"
                result_val[0]['snapshots'] = r_i.get_snapshots()
                connector_data = self.get_connector_data()
                result_val[0]['autoRemediate'] = connector_data.get("autoRemediate", False)
                if result_val[0]['snapshots']:
                    snapshot = result_val[0]['snapshots'][0]
                    logger.critical('\t\tSNAPSHOTID: %s', snapshot['id'])
                    logger.critical('\t\tPATHS: ')
                    for path in snapshot.get('paths', []):
                        logger.critical('\t\t\t %s', path)
                if not result:
                    logger.critical('\t\tLHS: %s', l_val)
                    logger.critical('\t\tRHS: %s', r_val)
                self.log_result(result_val[0])
        else:
            result_val[0].update({
                "result": "skipped",
                "reason": "Unsupported testcase format"
            })
        
        if 'dirpath' in self.testcase:
            del self.testcase['dirpath']
        return result_val


class ComparatorV02(ComparatorV01):
    """
    Override the validate method to run the comparisons
    """
    def __init__(self, container, dbname, collection_data, testcase, excludedTestIds, includeTests):
        ComparatorV01.__init__(self, container, dbname, collection_data, testcase, excludedTestIds, includeTests)

    def validate(self):
        return ComparatorV01.validate(self)


def main(container):
    from processor.connector.validation import run_container_validation_tests_filesystem
    try:
        run_container_validation_tests_filesystem(container)
        return True
    except Exception as ex:
        print("Exception: %s" % ex)
        return False


if __name__ == '__main__':
    import sys
    if len(sys.argv) > 1:
        main(sys.argv[1])
    else:        
        comparator = Comparator('0.1', 'validator', {}, {"testId": "4", "snapshotId1": "1", "attribute": "id", "comparison": "gt0 10"})
        val = comparator.validate()
        print(val)



"""
Performs all in built collection functions which are supported by terraform processor
"""
from processor.logging.log_handler import getlogger
import decimal
import itertools

logger = getlogger()

def element(ele, index):
    "return the element at given index from the list"
    if not ele:
        return
    return ele[index % len(ele)]

def length(ele):
    "return length of given element"
    return len(ele) if ele is not None else None

def chunklist(ele_list, n):
    "splits a single list into fixed-size chunks, returning a list of lists."
    def chunk(ele_list, n):
        for i in range(0, len(ele_list), n):  
            yield ele_list[i:i + n]
    return list(chunk(ele_list, n))    

def concat(*argv):
    "takes two or more lists and combines them into a single list"
    concat_list = []
    for arg in argv:  
        concat_list += arg
    return concat_list

def coalesce(*argv):
    "takes any number of arguments and returns the first one that isn't null or an empty string"
    for arg in argv:  
        if arg:
            return arg

def coalescelist(*argv):
    "takes any number of list arguments and returns the first one that isn't empty."
    return coalesce(*argv)

def compact(ele_list):
    "takes a list of strings and returns a new list with any empty string elements removed"
    return [ele for ele in ele_list if ele]

def distinct(ele_list):
    "takes a list and returns a new list with any duplicate elements removed"
    return list(set(ele_list))

def index(ele_list, value):
    "finds the element index for a given value in a list"
    return ele_list.index(value) if value in ele_list else -1

def lookup(json_data, find_key, default=None):
    "retrieves the value of a single element from a map. If the given key does not exist, return the default value."
    return json_data.get(find_key, default) if json_data and isinstance(json_data, dict) else default

def contains(ele_list, search_ele):
    "check element contains in a list or not"
    return True if search_ele in ele_list else False

def keys(json_data):
    "return the list of keys contains in given json"
    return list(json_data.keys())

def to_list(*args):
    "convert given arguments into list"
    return list(args)

def to_map(*args): 
    "convert given arguments into map"
    it = iter(args) 
    res_dict = dict(zip(it, it)) 
    return res_dict

def merge(*args):
    "merge more then one dict into single dict"
    res_dict = {}
    for arg in args:
        res_dict.update(arg)
    return res_dict

def reverse(ele_list):
    " takes the list as an argument and returns the list with reverse order "
    ele_list.reverse() 
    return ele_list  

def setintersection(*args):
    " takes multiple lists and produces a single list containing common elements "
    sets = iter(map(set, args))
    result = sets.next()
    for s in sets:
        result = result.intersection(s)
    return result

def to_range(start, limit=None, step=None):
    """ returns the list of numbers based on given start, limit and step values """
    def float_range(start, limit, step):
        while start < limit:
            yield float(start)
            start += decimal.Decimal(step)

    if isinstance(step, float):
        return [r for r in float_range(start, limit, step)]
    elif limit and limit < start and not step:
        return [r for r in range(start, limit, -1)]
    else:
        range_params = list(filter(None, [start, limit, step]))
        return [r for r in range(*range_params)]

def setintersection(*args):
    """ returns the list of common elements from given sets """
    intersection_list = []
    if args:
        intersection_list = list(set(args[0]).intersection(*args))
    return intersection_list
        
def setproduct(*args):
    """ creates a list with all possible combinations of elements of given sets """
    product_list= []
    if args:
        product_list = [list(ele) for ele in list(itertools.product(*args))]
    return product_list


from processor.logging.log_handler import getlogger
from processor.templates.base.template_parser import TemplateParser
from processor.helper.file.file_utils import exists_file,exists_dir


logger = getlogger()


class HelmTemplateParser(TemplateParser):
    def __init__(self, template_file, tosave=False, **kwargs):
        """
        Base Parser class for parse helm
        """
        super().__init__(template_file, tosave=False, **kwargs)
        self.type = {}

    def parse(self,file_path):
        return ""
    
    def validate(self,file_path):
        helm_source = file_path.rpartition("/")[0]
        check_file_path = "%s/Chart.yaml" % helm_source
        valeus_file_path = "%s/values.yaml" % helm_source
        template_dir_path = "%s/templates" % helm_source

        if  all([exists_file(check_file_path),exists_file(valeus_file_path),exists_dir(template_dir_path)]):
            return True
        return False


import json
import re
import collections
import traceback
import math

from processor.logging.log_handler import getlogger
logger = getlogger()


def shannon_entropy(data):
    if not data:
        return 0
    entropy = 0
    normalized_ent = 0
    n = 0
    for x in range(256):
        p_x = float(data.count(chr(x)))/len(data)
        if p_x > 0:
            n += 1
            entropy += - p_x*math.log(p_x, 2)

    if math.log(n) > 0:
        normalized_ent = entropy / math.log(n, 2)
    return entropy, normalized_ent


def get_paths(source):
    paths = []
    if isinstance(source, collections.MutableMapping):
        for k, v in source.items():
            paths.append([k])
            paths += [[k] + x for x in get_paths(v)]
    elif isinstance(source, collections.Sequence) and not isinstance(source, str):
        for i, v in enumerate(source):
            paths.append([i])
            paths += [[i] + x for x in get_paths(v)]
    return paths


def secret_finder(snapshot, PASSWORD_VALUE_RE, PASSWORD_KEY_RE=None, EXCLUDE_RE=None, shannon_entropy_password=False):
    output = {}
    errors = []
    try:
        issue_found = False
        skipped = True
        if isinstance(snapshot.get("resources"), list):
            for resource in snapshot.get("resources"):
                skipped = False
                path_list = get_paths(resource)
                for path in path_list:
                    nested_resource = resource
                    for key in path:
                        nested_resource = nested_resource[key]

                        if isinstance(nested_resource, (int, float, complex, bool)):
                            nested_resource = str(nested_resource)

                        if isinstance(nested_resource, str) and re.match(PASSWORD_VALUE_RE, nested_resource) and (re.match(PASSWORD_KEY_RE, str(key), re.I) if PASSWORD_KEY_RE else True) and (not(re.match(EXCLUDE_RE, str(nested_resource))) if EXCLUDE_RE else True):
                            if shannon_entropy_password:
                                _, normalized_entropy = shannon_entropy(
                                    nested_resource)
                                if normalized_entropy > 0.965:
                                    errors.append({
                                        "leaked_password_path": "resources/"+resource.get("type")+"/" + "/".join([str(path) for path in path]),
                                    })
                                    issue_found = True
                                    logger.warning("Leaked Password at:%s\nvalue:%s" % (
                                        "resources/"+resource.get("type")+"/" + "/".join([str(path) for path in path]), nested_resource))
                            else:
                                issue_found = True
                                errors.append({
                                    "leaked_password_path": "resources/"+resource.get("type")+"/" + "/".join([str(path) for path in path]),
                                })
                                logger.warning("Leaked Password at:%s\nvalue:%s" % (
                                    "resources/"+resource.get("type")+"/" + "/".join([str(path) for path in path]), nested_resource))

        output["issue"] = True if issue_found else False

        if errors:
            output["errors"] = errors

        output["skipped"] = skipped
        return output
    except Exception as ex:
        print(traceback.format_exc())
        output["issue"] = None
        output["err"] = str(ex)
        output["skipped"] = skipped
        return output


def google_password_leak(generated_snapshot: dict, kwargs={}) -> dict:

    PASSWORD_KEY_RE = r".*(?i)(password|secret).*"
    PASSWORD_VALUE_RE = r'^(?=^(?!\$\{.*\}$))(?=(?=.*[a-z][A-Z])|(?=.*[A-Z][a-z])|(?=.*[a-z][0-9])|(?=.*[0-9][a-z])|(?=.*[0-9][A-Z])|(?=.*[A-Z][0-9]))(.*[\^$*.\[\]{}\(\)?\-"!@\#%&\/,><\’:;|_~`]?)\S{8,99}$'
    EXCLUDE_REGEX = r'(?=.*([\[{(<]){1,})((\(.*\)){0,})((\[.*\]){0,})((\{.*\}){0,})((\<.*\>){0,})'

    output = secret_finder(
        generated_snapshot, PASSWORD_VALUE_RE, PASSWORD_KEY_RE, EXCLUDE_RE=EXCLUDE_REGEX)

    if output["issue"] == True:
        output["google_password_leak"] = "Ensure no hardcoded password set in the template"

    elif output["issue"] == None:
        output["google_password_leak"] = output["err"]
        output.pop("err")

    elif output["issue"] == False:
        output["google_password_leak"] = ""
    return output


def entropy_password(generated_snapshot: dict, kwargs={}) -> dict:

    PASSWORD_VALUE_RE = r'^(?=^(?!\$\{.*\}$))(?=(?=.*[a-z][A-Z])|(?=.*[A-Z][a-z])|(?=.*[a-z][0-9])|(?=.*[0-9][a-z])|(?=.*[0-9][A-Z])|(?=.*[A-Z][0-9]))(?=.*[^A-Za-z0-9])\S{8,99}$'
    EXCLUDE_CONTAINS = ['gcloud', 'access-approval', 'Overview', 'requests', 'approve', 'dismiss', 'get', 'list', 'settings', 'delete', 'update', 'access-context-manager', 'cloud-bindings', 'create', 'describe', 'levels', 'conditions', 'replace-all', 'perimeters', 'dry-run', 'drop', 'enforce', 'enforce-all', 'policies', 'active-directory', 'domains', 'describe-ldaps-settings', 'get-iam-policy', 'reset-admin-password', 'set-iam-policy', 'trusts', 'validate-state', 'update-ldaps-settings', 'operations', 'cancel', 'custom-jobs', 'stream-logs', 'endpoints', 'deploy-model', 'explain', 'predict', 'undeploy-model', 'hp-tuning-jobs', 'model-monitoring-jobs', 'pause', 'resume', 'models', 'upload', 'ai-platform', 'jobs', 'submit', 'prediction', 'training', 'local', 'train', 'add-iam-policy-binding', 'remove-iam-policy-binding', 'wait', 'versions', 'set-default', 'alpha', 'sql-integrations', 'peerings', 'local-run', 'raw-predict', 'index-endpoints', 'deploy-index', 'undeploy-index', 'indexes', 'tensorboard-experiments', 'tensorboard-runs', 'tensorboard-time-series', 'read', 'tensorboards', 'locations', 'anthos', 'apply', 'auth', 'login', 'config', 'controller', 'get-credentials', 'create-login-config', 'export', 'api-gateway', 'api-configs', 'apis', 'gateways', 'apigee', 'deploy', 'undeploy', 'applications', 'archives', 'deployments', 'developers', 'environments', 'organizations', 'provision', 'products', 'app', 'domain-mappings', 'ssl-certificates', 'artifacts', 'apt', 'import', 'docker', 'images', 'tags', 'add', 'packages', 'print-settings', 'gradle', 'mvn', 'npm', 'python', 'yum', 'repositories', 'asset', 'feeds', 'get-history', 'assured', 'workloads', 'activate-service-account', 'configure-docker', 'print-access-token', 'print-identity-token', 'revoke', 'bigtable', 'app-profiles', 'backups', 'clusters', 'hot-tablets', 'instances', 'tables', 'restore', 'upgrade', 'billing', 'accounts', 'projects', 'link', 'unlink', 'budgets', 'bms', 'datasets', 'copy', 'insert', 'show-rows', 'builds', 'configure', 'gke', 'enterprise-config', 'bitbucketserver', 'github', 'log', 'reject', 'triggers', 'cloud-source-repositories', 'pubsub', 'webhook', 'run', 'worker-pools', 'certificate-manager', 'certificates', 'dns-authorizations', 'maps', 'entries', 'cloud-shell', 'get-mount-command', 'scp', 'ssh', 'code', 'clean-up', 'dev', 'composer', 'check-upgrade', 'list-packages', 'list-upgrades', 'restart-web-server', 'storage', 'dags', 'data', 'plugins', 'compute', 'accelerator-types', 'addresses', 'backend-buckets', 'add-signed-url-key', 'delete-signed-url-key', 'backend-services', 'add-backend', 'edit', 'get-health', 'remove-backend', 'set-security-policy', 'update-backend', 'commitments', 'create-license', 'update-reservations', 'config-ssh', 'connect-to-serial-port', 'copy-files', 'diagnose', 'export-logs', 'routes', 'sosreport', 'disk-types', 'disks', 'add-labels', 'add-resource-policies', 'move', 'remove-labels', 'remove-resource-policies', 'resize', 'snapshot', 'external-vpn-gateways', 'firewall-policies', 'associations', 'clone-rules', 'list-rules', 'rules', 'firewall-rules', 'forwarding-rules', 'set-target', 'future-reservations', 'health-checks', 'grpc', 'http', 'http2', 'https', 'ssl', 'tcp', 'http-health-checks', 'https-health-checks', 'deprecate', 'describe-from-family', 'diff', 'vulnerabilities', 'describe-note', 'instance-groups', 'get-named-ports', 'list-instances', 'managed', 'abandon-instances', 'create-instance', 'delete-instances', 'describe-instance', 'export-autoscaling', 'instance-configs', 'list-errors', 'recreate-instances', 'resume-instances', 'rolling-action', 'replace', 'restart', 'start-update', 'stop-proactive-update', 'set-autohealing', 'set-autoscaling', 'set-instance-template', 'set-named-ports', 'set-standby-policy', 'set-target-pools', 'start-instances', 'stop-autoscaling', 'stop-instances', 'suspend-instances', 'update-autoscaling', 'update-instances', 'wait-until', 'wait-until-stable', 'unmanaged', 'add-instances', 'remove-instances', 'instance-templates', 'create-with-container', 'add-access-config', 'add-metadata', 'add-tags', 'attach-disk', 'bulk', 'delete-access-config', 'detach-disk', 'get-guest-attributes', 'get-serial-port-output', 'get-shielded-identity', 'network-interfaces', 'get-effective-firewalls', 'ops-agents', 'os-inventory', 'remove-metadata', 'remove-tags', 'reset', 'send-diagnostic-interrupt', 'set-disk-auto-delete', 'set-machine-type', 'set-min-cpu-platform', 'set-name', 'set-scheduling', 'set-scopes', 'simulate-maintenance-event', 'start', 'stop', 'suspend', 'tail-serial-port-output', 'update-access-config', 'update-container', 'update-from-file', 'instant-snapshots', 'interconnects', 'attachments', 'dedicated', 'partner', 'get-diagnostics', 'macsec', 'add-key', 'get-config', 'remove-key', 'machine-images', 'machine-types', 'network-edge-security-services', 'network-endpoint-groups', 'list-network-endpoints', 'network-firewall-policies', 'networks', 'list-ip-addresses', 'list-ip-owners', 'list-routes', 'subnets', 'expand-ip-range', 'list-usable', 'org-security-policies', 'copy-rules', 'os-config', 'guest-policies', 'lookup', 'instance-os-policies-compliances', 'inventories', 'os-policy-assignments', 'list-revisions', 'os-upgrade', 'patch-deployments', 'patch-jobs', 'execute', 'list-instance-details', 'vulnerability-reports', 'os-login', 'describe-profile', 'remove-profile', 'ssh-keys', 'remove', 'packet-mirrorings', 'project-info', 'set-default-service-account', 'set-usage-bucket', 'public-advertised-prefixes', 'public-delegated-prefixes', 'delegated-sub-prefixes', 'regions', 'reservations', 'reset-windows-password', 'resource-policies', 'group-placement', 'instance-schedule', 'snapshot-schedule', 'vm-maintenance', 'concurrency-limit', 'maintenance-window', 'create-snapshot-schedule', 'create-vm-maintenance', 'routers', 'add-bgp-peer', 'add-interface', 'get-nat-mapping-info', 'get-status', 'nats', 'remove-bgp-peer', 'remove-interface', 'update-bgp-peer', 'update-interface', 'security-policies', 'list-preconfigured-expression-sets', 'service-attachments', 'shared-vpc', 'associated-projects', 'disable', 'enable', 'get-host-project', 'list-associated-resources', 'list-host-projects', 'sign-url', 'snapshots', 'sole-tenancy', 'node-groups', 'list-nodes', 'node-templates', 'node-types', 'ssl-policies', 'list-available-features', 'start-iap-tunnel', 'target-grpc-proxies', 'target-http-proxies', 'target-https-proxies', 'target-instances', 'target-pools', 'add-health-checks', 'remove-health-checks', 'set-backup', 'target-ssl-proxies', 'target-tcp-proxies', 'target-vpn-gateways', 'tpus', 'execution-groups', 'reimage', 'tpu-vm', 'service-identity', 'url-maps', 'add-host-rule', 'add-path-matcher', 'invalidate-cdn-cache', 'list-cdn-cache-invalidations', 'remove-host-rule', 'remove-path-matcher', 'set-default-service', 'validate', 'vpn-gateways', 'vpn-tunnels', 'zones', 'configurations', 'activate', 'set', 'unset', 'container', 'aws', 'get-kubeconfig', 'get-server-config', 'node-pools', 'azure', 'clients', 'get-public-cert', 'backup-restore', 'backup-plans', 'restores', 'volume-backups', 'volume-restores', 'binauthz', 'attestations', 'sign-and-create', 'attestors', 'public-keys', 'continuous-validation', 'create-signature-payload', 'policy', 'export-system-policy', 'create-auto', 'hub', 'cloudrun', 'config-management', 'fetch-for-apply', 'status', 'unmanage', 'version', 'features', 'identity-service', 'ingress', 'memberships', 'generate-gateway-rbac', 'register', 'unregister', 'mesh', 'multi-cluster-services', 'service-directory', 'add-tag', 'list-tags', 'untag', 'rollback', 'data-catalog', 'crawler-runs', 'crawlers', 'entry-groups', 'search', 'tag-templates', 'fields', 'enum-values', 'rename', 'taxonomies', 'policy-tags', 'database-migration',
                        'connection-profiles', 'cloudsql', 'mysql', 'migration-jobs', 'generate-ssh-script', 'promote', 'verify', 'dataflow', 'export-steps', 'resume-unsupported-sdk', 'logs', 'metrics', 'dataproc', 'autoscaling-policies', 'create-from-file', 'kill', 'hadoop', 'hive', 'pig', 'presto', 'pyspark', 'spark', 'spark-r', 'spark-sql', 'workflow-templates', 'add-job', 'instantiate', 'instantiate-from-file', 'remove-dag-timeout', 'remove-job', 'set-cluster-selector', 'set-dag-timeout', 'set-managed-cluster', 'datastore', 'databases', 'cleanup', 'delivery-pipelines', 'releases', 'rollouts', 'targets', 'deployment-manager', 'cancel-preview', 'manifests', 'resources', 'type-providers', 'types', 'dialogflow', 'agent', 'query', 'entity-types', 'intents', 'dlp', 'datasources', 'bigquery', 'analyze', 'inspect', 'gcs', 'redact', 'job-triggers', 'text', 'dns', 'active-peering-zones', 'dns-keys', 'managed-zones', 'record-sets', 'changes', 'transaction', 'abort', 'response-policies', 'registrations', 'authorization-code', 'contacts', 'management', 'get-register-parameters', 'get-transfer-parameters', 'search-domains', 'transfer', 'emulators', 'env-init', 'env-unset', 'firestore', 'spanner', 'configs', 'quota', 'services', 'check-iam-policy', 'undelete', 'essential-contacts', 'filestore', 'firebase', 'test', 'android', 'locales', 'ios', 'ip-blocks', 'network-profiles', 'composite', 'functions', 'add-invoker-policy-binding', 'call', 'event-types', 'remove-invoker-policy-binding', 'game', 'servers', 'describe-rollout', 'fetch-state', 'update-rollout', 'realms', 'genomics', 'pipelines', 'healthcare', 'annotation-stores', 'evaluate', 'consent-stores', 'check-data-access', 'evaluate-user-consents', 'query-accessible-data', 'deidentify', 'dicom-stores', 'fhir-stores', 'hl7v2-stores', 'nlp', 'analyze-entities', 'help', 'iam', 'list-grantable-roles', 'list-testable-permissions', 'lint-condition', 'roles', 'service-accounts', 'keys', 'get-public-key', 'sign-blob', 'sign-jwt', 'workload-identity-pools', 'create-cred-config', 'providers', 'create-aws', 'create-oidc', 'update-aws', 'update-oidc', 'iap', 'oauth-brands', 'oauth-clients', 'reset-secret', 'web', 'identity', 'groups', 'check-transitive-membership', 'get-membership-graph', 'modify-membership-roles', 'search-transitive-groups', 'search-transitive-memberships', 'preview', 'ids', 'init', 'interactive', 'iot', 'devices', 'commands', 'send', 'get-value', 'credentials', 'clear', 'bind', 'list-bound-devices', 'unbind', 'states', 'registries', 'kms', 'asymmetric-decrypt', 'asymmetric-sign', 'decrypt', 'encrypt', 'import-jobs', 'keyrings', 'remove-rotation-schedule', 'set-primary-version', 'set-rotation-schedule', 'destroy', 'get-certificate-chain', 'mac-sign', 'mac-verify', 'lifesciences', 'logging', 'buckets', 'resource-descriptors', 'sinks', 'tail', 'views', 'write', 'media', 'memcache', 'apply-parameters', 'apply-software-update', 'metastore', 'imports', 'language', 'analyze-entity-sentiment', 'analyze-sentiment', 'analyze-syntax', 'classify-text', 'speech', 'recognize', 'recognize-long-running', 'translate', 'batch-translate-text', 'detect-language', 'get-supported-languages', 'translate-text', 'video', 'detect-explicit-content', 'detect-labels', 'detect-object', 'detect-shot-changes', 'detect-text', 'transcribe-speech', 'vision', 'detect-document', 'detect-faces', 'detect-image-properties', 'detect-landmarks', 'detect-logos', 'detect-objects', 'detect-product', 'detect-safe-search', 'detect-text-pdf', 'detect-text-tiff', 'detect-web', 'product-search', 'product-sets', 'add-product', 'list-products', 'remove-product', 'delete-all', 'suggest-crop', 'ml-engine', 'monitoring', 'channel-descriptors', 'channels', 'dashboards', 'network-connectivity', 'hubs', 'spokes', 'network-security', 'authorization-policies', 'client-tls-policies', 'server-tls-policies', 'network-services', 'endpoint-policies', 'grpc-routes', 'http-filters', 'http-routes', 'observability-policies', 'tcp-routes', 'notebooks', 'is-upgradeable', 'policy-troubleshoot', 'get-ancestors', 'publicca', 'external-account-keys', 'lite-operations', 'lite-reservations', 'list-topics', 'lite-subscriptions', 'ack-up-to', 'seek', 'subscribe', 'lite-topics', 'list-subscriptions', 'publish', 'schemas', 'validate-message', 'validate-schema', 'subscriptions', 'ack', 'modify-ack-deadline', 'modify-message-ack-deadline', 'modify-push-config', 'pull', 'topics', 'detach-subscription', 'list-snapshots', 'recaptcha', 'migrate', 'recommender', 'insights', 'mark-accepted', 'mark-active', 'mark-dismissed', 'recommendations', 'mark-claimed', 'mark-failed', 'mark-succeeded', 'recommender-configs', 'redis', 'failover', 'get-auth-string', 'reschedule-maintenance', 'resource-config', 'bulk-export', 'list-resource-types', 'terraform', 'generate-import', 'resource-manager', 'folders', 'liens', 'org-policies', 'allow', 'deny', 'disable-enforce', 'enable-enforce', 'set-policy', 'bindings', 'values', 'resource-settings', 'list-values', 'set-value', 'unset-value', 'revisions', 'update-traffic', 'scc', 'assets', 'get-parent', 'get-project', 'group', 'list-marks', 'run-discovery', 'update-marks', 'findings', 'set-mute', 'notifications', 'describe-explicit', 'modules', 'sources', 'scheduler', 'app-engine', 'namespaces', 'resolve', 'api-keys', 'clone', 'get-key-string', 'peered-dns-domains', 'vpc-peerings', 'connect', 'disable-vpc-service-controls', 'enable-vpc-service-controls', 'source', 'project-configs', 'repos', 'update-metadata', 'ddl', 'execute-sql', 'sessions', 'rows', 'sql', 'patch', 'bak', 'csv', 'flags', 'promote-replica', 'reset-ssl-config', 'restore-backup', 'client-certs', 'server-ca-certs', 'rotate', 'ssl-certs', 'tiers', 'users', 'set-password', 'survey', 'tasks', 'acknowledge', 'cancel-lease', 'create-app-engine-task', 'create-http-task', 'create-pull-task', 'lease', 'queues', 'create-app-engine-queue', 'create-pull-queue', 'purge', 'update-app-engine-queue', 'update-pull-queue', 'renew-lease', 'trace', 'transcoder', 'templates', 'vmware', 'nodetypes', 'nsx', 'privateclouds', 'vcenter', 'web-security-scanner', 'scan-configs', 'scan-runs', 'list-crawled-urls', 'browse', 'test-ip', 'disable-debug', 'enable-debug', 'open-console', 'set-traffic', 'get-operation', 'list-vulnerabilities', 'scan', 'analyze-iam-policy', 'analyze-iam-policy-longrunning', 'analyze-move', 'search-all-iam-policies', 'search-all-resources', 'application-default', 'set-quota-project', 'beta', 'gen-config', 'migrate-config', 'cron-xml-to-yaml', 'datastore-indexes-xml-to-yaml', 'dispatch-xml-to-yaml', 'queue-xml-to-yaml', 'repair', 'export-iam-policy-analysis', 'get-screenshot', 'vpc-access', 'connectors', 'data-fusion', 'flex-template', 'build', 'drain', 'show', 'datastream', 'discover', 'fetch-static-ips', 'private-connections', 'streams', 'debug', 'logpoints', 'gen-repo-info-file', 'list-user-verified', 'error-reporting', 'events', 'report', 'eventarc', 'attributes', 'method-names', 'service-names', 'brokers', 'network-management', 'connectivity-tests', 'rerun', 'privateca', 'reusable-configs', 'roots', 'subordinates', 'get-csr', 'runtime-config', 'variables', 'watch', 'waiters', 'secrets', 'replication', 'access', 'workflows', 'executions', 'describe-last', 'wait-last', 'cheat-sheet', 'components', 'install', 'reinstall', 'set-service-account', 'postgresql', 'feedback', 'simulator', 'replay-recent-access', 'info', 'policy-intelligence', 'query-activity', 'pools', 'get-ca-certs', 'replicate', 'get-ancestors-iam-policy', 'topic', 'accessibility', 'arg-files', 'cli-trees', 'client-certificate', 'command-conventions', 'datetimes', 'escaping', 'filters', 'flags-file', 'formats', 'gcloudignore', 'offline-help', 'projections', 'resource-keys', 'startup', 'uninstall', 'workspace-add-ons', 'install-status', 'get-authorization', 'sha', 'sha256']
    EXCLUDE_REGEX = [
        "(?=^([a-zA-Z0-9]+\.+[a-zA-Z0-9]+\.[a-zA-Z0-9]+)(?![A-Za-z0-9])$)",
        "(?=^([a-zA-Z0-9]+\.+[a-zA-Z0-9]+\.[a-zA-Z0-9]+\.[a-zA-Z0-9]+)(?![A-Za-z0-9])$)",
        "(?=^([a-zA-Z0-9]+\_+[a-zA-Z0-9]+\_[a-zA-Z0-9]+)(?![A-Za-z0-9])$)",
        "(?=^([a-zA-Z0-9]+\-+[a-zA-Z0-9]+\-[a-zA-Z0-9]+)(?![A-Za-z0-9])$)",
        "(?=^([a-zA-Z0-9]+\.+[a-zA-Z0-9]*)$)",
        "(?=^(\/+\w{0,}){0,}$)",
        "(?=^([a-zA-Z0-9]*_[a-zA-Z0-9]*)$)",
        "(?=^([a-zA-Z0-9]*-[a-zA-Z0-9]*)$)",
    ]
    EXCLUDE_STARTSWITH = [
        "arn:"
    ]

    exclude_contains_regex = "(?=.*(?i)(%s).*)" % ("|".join(EXCLUDE_CONTAINS))

    exclude_startswith_regex = "(?=(%s))" % (
        "|".join(["^"+exclude for exclude in EXCLUDE_STARTSWITH]))
    exclude_regex = "|".join(EXCLUDE_REGEX)

    combined_exclude_regex = "|".join(
        [exclude_contains_regex, exclude_startswith_regex, exclude_regex])
    combined_exclude_regex = re.compile(combined_exclude_regex)

    output = secret_finder(
        generated_snapshot, PASSWORD_VALUE_RE, PASSWORD_KEY_RE=None, EXCLUDE_RE=combined_exclude_regex, shannon_entropy_password=True)

    if output["issue"] == True:
        output["entropy_password_err"] = "There is a possibility that a value might contains a secret string or password"

    elif output["issue"] == None:
        output["entropy_password_err"] = output["err"]
        output.pop("err")

    else:
        output["entropy_password_err"] = ""
    return output



import traceback
from ruamel.yaml import YAML
from processor.helper.file.file_utils import exists_file
from processor.logging.log_handler import getlogger
import re

logger = getlogger()

yaml = YAML()

class JinjaConverter:
    """ analyse, change and revert YAML/jinja2 mixture to/from valid YAML"""
    def __init__(self):
        self.property_start = '""#<<<'
        self.property_end = '>>>#'
        self.comment_start = "#>"
        self.comment_end = "<#"
        self.block_comment_start = "#{#"
        self.quote = "<quotes>"
        self.double_quote = "<dquotes>"

    def replace_newline_character(self, replace_string, revert=False):
        if revert:
            return replace_string.replace("\n#", "\n")
        else:
            return "\n#".join(replace_string.splitlines())

    def comment_jinja_syntax(self, s):
        """ comment the jinja syntax and return the string with replaced values """

        s = re.sub("({#[\S\s]*?#})", lambda m: self.replace_newline_character(m.group(0)), s, flags=re.DOTALL)
        updated_string = s.replace('{{ ', self.property_start) \
            .replace(' }}', self.property_end) \
            .replace('{%', self.comment_start) \
            .replace('%}', self.comment_end) \
            .replace('"', self.double_quote) \
            .replace("'", self.quote) \
            .replace("{#", self.block_comment_start)
        return updated_string

    def revert(self, s):
        """ reverts the string to it's original states and retuns the updated string """
        if s and isinstance(s, str):
            new_string = s.replace(self.double_quote, '"') \
                .replace(self.quote, "'") \
                .replace(self.property_start, '{{ ') \
                .replace(self.property_end, ' }}') \
                .replace(self.comment_start, '{%') \
                .replace(self.comment_end, '%}') \
                .replace(self.block_comment_start, "{#")
            
            new_string = re.sub("({#[\S\s]*?#})", lambda m: self.replace_newline_character(m.group(0), revert=True), new_string, flags=re.DOTALL)    
            return new_string
        return s
        
    def jinja_to_json(self, file_name, transform=False):
        """ convert jinja file to json object without processing variables """
        json_data = None
        try:
            with open(file_name) as fp:
                if transform:
                    json_data = yaml.load(self.comment_jinja_syntax(fp.read()))
                else:
                    json_data = yaml.load(fp.read())
        except Exception as e:
            logger.info("Failed to convert jinja template into json object %s ", str(e))
        return json_data

    def save_json_to_jinja_file(self, json_data, output_file, transform=False):
        """ convert Jinja file to json object """
        try:
            if exists_file(output_file):
                with open(output_file, 'w') as fp:
                    if transform:
                        yaml.dump(json_data, fp, transform=self.revert)
                    else:
                        yaml.dump(json_data, fp)
                    return True
            logger.info("File doesnot exist at given path : %s", output_file)
        except:
            logger.info("Failed to save json data into jinja file")
            logger.error(traceback.format_exc())
        return False


from processor.templates.terraform.helper.function import collection_functions
from processor.templates.terraform.helper.function import numeric_functions
from processor.templates.terraform.helper.function import string_functions

default_functions = [
    # collection functions
    { "expression" : "(element[(].*[,].*[)])$", "method" : collection_functions.element},
    { "expression" : "(length[(].*[)])$", "method" : collection_functions.length},
    { "expression" : "(chunklist[(].*[)])$", "method" : collection_functions.chunklist},
    { "expression" : "(concat[(].*[)])$", "method" : collection_functions.concat},
    { "expression" : "(coalesce[(].*[)])$", "method" : collection_functions.coalesce},
    { "expression" : "(coalescelist[(].*[)])$", "method" : collection_functions.coalescelist},
    { "expression" : "(compact[(].*[)])$", "method" : collection_functions.compact},
    { "expression" : "(distinct[(].*[)])$", "method" : collection_functions.distinct},
    { "expression" : "(index[(].*[)])$", "method" : collection_functions.index},
    { "expression" : "(lookup[(].*[,].*[)])$", "method" : collection_functions.lookup},
    { "expression" : "(contains[(].*[,].*[)])$", "method" : collection_functions.contains},
    { "expression" : "(keys[(].*[)])$", "method" : collection_functions.keys},
    { "expression" : "(list[(].*[)])$", "method" : collection_functions.to_list},
    { "expression" : "(map[(].*[)])$", "method" : collection_functions.to_map},
    { "expression" : "(merge[(].*[)])$", "method" : collection_functions.merge},
    { "expression" : "(reverse[(].*[)])$", "method" : collection_functions.reverse},
    { "expression" : "(range[(].*[)])$", "method" : collection_functions.to_range},
    { "expression" : "(setintersection[(].*[)])$", "method" : collection_functions.setintersection},
    { "expression" : "(setproduct[(].*[)])$", "method" : collection_functions.setproduct},

    # numeric functions
    { "expression" : "(abs[(].*[)])$", "method" : numeric_functions.to_abs},
    { "expression" : "(ceil[(].*[)])$", "method" : numeric_functions.ceil},
    { "expression" : "(floor[(].*[)])$", "method" : numeric_functions.floor},
    { "expression" : "(log[(].*[)])$", "method" : numeric_functions.log},
    { "expression" : "(max[(].*[)])$", "method" : numeric_functions.to_max},
    { "expression" : "(min[(].*[)])$", "method" : numeric_functions.to_min},
    { "expression" : "(pow[(].*[)])$", "method" : numeric_functions.pow},
    { "expression" : "(signum[(].*[)])$", "method" : numeric_functions.signum},

    # string funtion
    { "expression" : "(chomp[(].*[)])$", "method" : string_functions.chomp },
    { "expression" : "(join[(].*[)])$", "method" : string_functions.join },
    { "expression" : "(lower[(].*[)])$", "method" : string_functions.lower },
    { "expression" : "(replace[(].*[)])$", "method" : string_functions.replace },
    { "expression" : "(split[(].*[)])$", "method" : string_functions.split },
    { "expression" : "(trim[(].*[)])$", "method" : string_functions.trim },
    { "expression" : "(trimprefix[(].*[)])$", "method" : string_functions.trimprefix },
    { "expression" : "(trimsuffix[(].*[)])$", "method" : string_functions.trimsuffix },
    { "expression" : "(trimspace[(].*[)])$", "method" : string_functions.trimspace },
    { "expression" : "(upper[(].*[)])$", "method" : string_functions.upper },
    { "expression" : "(strrev[(].*[)])$", "method" : string_functions.strrev },
    { "expression" : "(substr[(].*[)])$", "method" : string_functions.substr },
    { "expression" : "(title[(].*[)])$", "method" : string_functions.title },
    { "expression" : "(format[(].*[)])$", "method" : string_functions.format },
]


import hcl
import hcl2
from lark import tree
from processor.helper.hcl import parser
from processor.logging.log_handler import getlogger

logger = getlogger()

def hcl_to_json(file_path):
    """
    converts the hcl file to json file
    """
    json_data = {}
    try:
        with open(file_path, 'r', encoding="utf-8") as fp:
            json_data = parser.loads(fp)
    except Exception as e:
        error = str(e)
        error = error.split("Expected one of")[0]
        logger.error("Unspported terraform file, error while parsing file: %s , error: %s", file_path, error)

    return json_data

if __name__ == "__main__":
    json_data = hcl_to_json("/tmp/extrasg.tf")
    import json
    print(json.dumps(json_data, indent=2))




"""Mongo db driver and utility functions."""
import os
import collections
from datetime import datetime, timedelta
from pymongo import MongoClient, TEXT, ASCENDING, DESCENDING
from pymongo.errors import ServerSelectionTimeoutError
from bson.objectid import ObjectId
from processor.helper.config.config_utils import config_value, DATABASE, DBNAME, DBURL
from processor.helper.config.rundata_utils import put_in_cachedata, get_from_cachedata
from processor.logging.dburl_kv import get_dburl


MONGO = None
COLLECTION = 'resources'
TIMEOUT = 3000
EXPIRE_TIME = 14400 # 4 hours
DBURL = "dburl"
DBURL_EXPIREATION = "dburl_expiration"

def mongoconnection(dbport=27017, to=TIMEOUT):
    """ Global connection handle for mongo """
    global MONGO
    if MONGO:
       return MONGO
    dburl = get_dburl_from_cache()
    # print("Dburl: %s" % dburl)
    if dburl:
        MONGO = MongoClient(host=dburl, serverSelectionTimeoutMS=to)
    else:
        MONGO = MongoClient(port=dbport, serverSelectionTimeoutMS=to)
    return MONGO

def get_dburl_from_cache():
    """ returns the database url from cache. if it is not preset in cache then it will return from vault """
    dburl = os.getenv('DBURL', None)

    if not dburl:
        memory_dburl = get_from_cachedata(DBURL)
        memory_dburl_expiration = get_from_cachedata(DBURL_EXPIREATION)

        now = datetime.now()
        current_timestamp = datetime.timestamp(now)

        if memory_dburl and memory_dburl_expiration:
            if current_timestamp > memory_dburl_expiration:
                dburl = get_dburl()
                expired_time = datetime.timestamp(now + timedelta(seconds=EXPIRE_TIME))
                put_in_cachedata(DBURL, dburl)
                put_in_cachedata(DBURL_EXPIREATION, expired_time)
            else:
                dburl = memory_dburl
        else:
            dburl = get_dburl()
            expired_time = datetime.timestamp(now + timedelta(seconds=EXPIRE_TIME))
            put_in_cachedata(DBURL, dburl)
            put_in_cachedata(DBURL_EXPIREATION, expired_time)

    return dburl

def mongodb(dbname=None):
    """ Get the dbhandle for the database, if none then default database, 'test' """
    dbconnection = mongoconnection()
    if dbname:
        # print("DB NAME: " + dbname)
        db = dbconnection[dbname]
    else:
        db = dbconnection['test']
    return db


def init_db():
    dbname = None
    try:
        dbconn = mongoconnection()
        _ = dbconn.list_database_names()
        dbname = config_value(DATABASE, DBNAME)
        create_indexes(COLLECTION, dbname, [('timestamp', TEXT)])
        db_init = True
    except ServerSelectionTimeoutError as ex:
        db_init = False
    return dbname, db_init


def get_collection(dbname, collection):
    """ Get the collection new or existing. """
    coll = None
    db = mongodb(dbname)
    if db and collection:
        coll = db[collection]
    return coll


def collection_names(dbname):
    """ Find all the collections in the databases. """
    db = mongodb(dbname)
    # colls = db.collection_names(include_system_collections=False)
    colls = db.list_collection_names()
    return colls


def sort_dict(data):
    vals = []
    for key, val in data.items():
        if isinstance(val, dict):
            sub_dict = sort_dict(val)
            vals.append((key, sub_dict))
        else:
            vals.append((key, val))
    sorted_vals = sorted(vals, key=lambda x: x[0])
    return collections.OrderedDict(sorted_vals)


def update_one_document(doc, collection, dbname):
    """ Update the document into the collection. """
    coll = get_collection(dbname, collection)
    if coll and doc:
        coll.save(doc)

def find_and_update_document(collection, dbname, query, update_value):
    """ find and update single document into the collection. """
    db = mongodb()
    collection = get_collection(dbname, collection)
    if collection:
        collection.update_many(query, update_value)
        return True
    return False

def insert_one_document(doc, collection, dbname, check_keys=True):
    """ Insert one document into the collection. """
    doc_id_str = None
    coll = get_collection(dbname, collection)
    if coll and doc:
        if check_keys:
            doc_id = coll.insert_one(sort_dict(doc))
            doc_id_str = str(doc_id.inserted_id)
        else:
            doc_ids = coll.insert(sort_dict(doc), check_keys=False)
            doc_id_str = str(doc_ids) if doc_ids else None
    return doc_id_str


def insert_documents(docs, collection, dbname):
    """ Insert multiple documents into the collection. """
    doc_ids = []
    coll = get_collection(dbname, collection)
    if coll and docs:
        result = coll.insert_many(docs)
        doc_ids = [str(inserted_id) for inserted_id in result.inserted_ids]
    return doc_ids

def delete_documents(collection, query, dbname):
    """ Delete the document based on the query """
    db = mongodb(dbname)
    collection = db[collection] if db and collection else None
    if collection:
        doc = collection.delete_many(query)
        return True
    return False


def check_document(collection, docid, dbname=None):
    """ Find the document based on the docid """
    doc = None
    db = mongodb(dbname)
    collection = db[collection] if db and collection else None
    if collection:
        obj_id = docid if isinstance(docid, ObjectId) else ObjectId(docid)
        doc = collection.find_one({'_id': obj_id})
    return doc


def get_documents(collection, query=None, dbname=None, sort=None, limit=10, skip=0, proj=None, _id=False):
    """ Find the documents based on the query """
    docs = None
    db = mongodb(dbname)
    collection = db[collection] if db and collection else None
    if collection:
        query = {} if query is None else query
        proj = proj if proj else {}
        if not _id:
            proj['_id'] = 0
        if sort:
            if proj:
                results = collection.find(filter=query, projection=proj).sort(sort).limit(limit).skip(skip)
            else:
                results = collection.find(filter=query).sort(sort).limit(limit).skip(skip)
        else:
            results = collection.find(query).limit(limit).skip(skip)
        docs = [result for result in results]
    return docs


def count_documents(collection, query=None, dbname=None):
    """ Count the documents based on the query """
    count = None
    db = mongodb(dbname)
    collection = db[collection] if db and collection else None
    if collection:
        query = {} if query is None else query
        count = collection.count_documents(query)
    return count


def distinct_documents(collection, field=None, dbname=None):
    """ Count the documents based on the query """
    count = []
    db = mongodb(dbname)
    collection = db[collection] if db and collection else None
    if collection and field:
        count = collection.distinct(field)
    return count

def create_indexes(collection, dbname, fields):
    """ The fields to be indexed """
    result = None
    db = mongodb(dbname)
    collection = db[collection] if db and collection else None
    if collection:
        # index_fields = [(field, ASCENDING) for field in fields]
        result = collection.create_index(fields, unique=True)
    return result

def get_collection_size(collection_name):
    db_name = config_value(DATABASE, DBNAME)
    collection = get_collection(db_name, collection_name)
    return collection.count()

def index_information(collection, dbname):
    """ index information of the collection """
    index_info = None
    db = mongodb(dbname)
    collection = db[collection] if db and collection else None
    if collection:
        index_info = sorted(list(collection.index_information()))
    return index_info

def sort_field(name, asc=True):
    return (name, ASCENDING if asc else DESCENDING)

# def main():
#   # dbname = 'business'
#   # coll = 'reviews'
#   # # mongodb(dbname)
#   # obj_str = "5bfe2aef7456213682bebaa6"
#   # doc = check_document(coll, obj_str, dbname)
#   # print(doc)
#   # obj_id = ObjectId(obj_str)
#   # doc = check_document(coll, obj_id, dbname)
#   # print(doc)
#   # doc = check_document(coll, None, dbname)
#   # print(doc)
#   # colls = collection_names(dbname)
#   # print(colls)
#   # doc = {'name': 'Test name', 'gender': True}
#   # newcoll = 'users'
#   # user_id = insert_one_document(doc, newcoll, dbname)
#   # print(user_id)
#   # doc = check_document(newcoll, user_id, dbname)
#   # print(doc)
#   # docs = [
#   #         {'name': 'Test1 name', 'gender': True},
#   #         {'name': 'Test2 name', 'gender': False}
#   #        ]
#   # doc_ids = insert_documents(docs, newcoll, dbname)
#   # print(doc_ids)
#   # colls = collection_names(dbname)
#   # print(colls)
#   # docs = get_documents(newcoll, None, dbname)
#   # print(docs)
#   # count = count_documents(newcoll, None, dbname)
#   # print(count)
#   # count = count_documents(newcoll, query={'gender': False}, dbname=dbname)
#   # print(count)
#   # print(index_information(coll, dbname))
#   # # print(create_indexes(newcoll, dbname, ['name']))
#   # print(index_information(newcoll, dbname))
#
#     a = {'a': 1, 'b': 2, 'f': 5, 'c': 3, 'd': 4}
#     b = {'z': a, 'y': {'x': 1, 'a': a}, 'm': 2, 'n': 'abc'}
#     c = { 'n': 'abc', 'y': {'x': 1, 'a': a}, 'z': a, 'm': 2}
#     d = sort_dict(b)
#     e = sort_dict(c)
#     print(d)
#     print(d.keys())
#     print(e)
#   d_str = json.dumps(d)
#   e_str = json.dumps(e)
#   is_match = True if d_str == e_str else False
#   print(b)
#   print(c)
#   print(hashlib.md5(json.dumps(b).encode('utf-8')).hexdigest())
#   print(hashlib.md5(json.dumps(c).encode('utf-8')).hexdigest())
#   print(d_str)
#   print(e_str)
#   print(is_match)
#
#
#
#
# if __name__ == "__main__":
#     main()




from processor.template_processor.aws_template_processor import AWSTemplateProcessor
from processor.template_processor.azure_template_processor import AzureTemplateProcessor
from processor.template_processor.google_template_processor import GoogleTemplateProcessor
from processor.template_processor.terraform_template_processor import TerraformTemplateProcessor
from processor.template_processor.kubernetes_template_processor import KubernetesTemplateProcessor
from processor.template_processor.yaml_template_processor import YamlTemplateProcessor
from processor.template_processor.json_template_processor import JsonTemplateProcessor
from processor.template_processor.helm_chart_template_processor import HelmChartTemplateProcessor
from processor.template_processor.ack_processor import AckTemplateProcessor
from processor.template_processor.aso_processor import AsoTemplateProcessor
from processor.template_processor.kcc_processor import KccTemplateProcessor
from processor.template_processor.base.base_template_processor import TemplateProcessor

TEMPLATE_NODE_TYPES = {
    "cloudformation": AWSTemplateProcessor,
    "arm" : AzureTemplateProcessor,
    "deploymentmanager" : GoogleTemplateProcessor,
    "terraform" : TerraformTemplateProcessor,
    "kubernetesObjectFiles" : KubernetesTemplateProcessor,
    "yaml" : YamlTemplateProcessor,
    "json": JsonTemplateProcessor,
    "helmChart" : HelmChartTemplateProcessor,
    "ack" : AckTemplateProcessor, # AWS Controllers for Kubernetes
    "aso" : AsoTemplateProcessor, # Azure Service Operator
    "kcc" : KccTemplateProcessor, # GCP Kubernetes Config Connector
    "common" : TemplateProcessor # TemplateProcessor
}